# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Pivot helper functions ."""
from collections import abc
from functools import wraps
from typing import Any, Callable, Dict, Optional, Tuple, Union
import warnings

import attr
import pandas as pd

from .._version import VERSION
from . import entities

__version__ = VERSION
__author__ = "Ian Hellen"


_DF_SRC_COL_PARAM_NAMES = [
    "column",
    "input_column",
    "input_col",
    "src_column",
    "src_col",
]


@attr.s(auto_attribs=True)
class PivotRegistration:
    """
    Pivot registration for function.

    Notes
    -----
    src_module : str
        The src_module to import
    src_class : str, optional
        class to import and instantiate that contains the function/method
        (not needed if the target function is a pure Python function)
    src_func_name: Callable
        The function to wrap.
    func_new_name: str, optional
        Rename the function to this, defaults to `src_func_name`
    input_type : str
        The input data type that the function is expecting.
        One of 'dataframe', 'iterable', 'value'
    can_iterate: bool, optional
        True if the function supports being called multiple times
        (for iterable input). Default is True
    entity_map: Dict[str, str]
        dict of entities supported (keys) and attribute to use from
        entity as input to the function
    func_df_param_name: str
        The name of the parameter that `func` takes the input value
        e.g. func(ip=my_address) => 'ip' == func_df_col_param_name.
        In the case of a DataFrame, this is usually 'data'
    func_df_col_param_name: str
        The name that the target function uses to identify the
        column to use for input in the input DataFrame.
    func_out_column_name: str, optional
        The name of the column in the output DF to use as a key to join
        to the input. If None, use `func_df_col_param_name`
    func_static_params: Optional[Dict[str, Any]]
        static parameters (kwargs) that are always passed
        to the target function
    func_input_value_arg: Optional[str]
        The name of kwarg passed to the function that contain
        the input value. If function supports DF input, `func_df_col_param_name`
        will be used and this is not needed.
    src_config_path : Optional[str]
        The source path that the configuration was read from, default None.
    src_config_entry : Optional[str]
        The entry name in the configuration file, default None.
    entity_container_name : Optional[str]
        The name of the container in the entity that will hold
        this pivot function.
    return_raw_output : bool
        Return raw output from the wrapped function, do not
        try to format into a DataFrame. Default is False.
    create_shortcut : bool
        If True, create a shortcut function directly on the entity.

    """

    input_type: str
    entity_map: Dict[str, str]
    func_df_param_name: Optional[str] = None
    func_out_column_name: Optional[str] = None
    func_df_col_param_name: Optional[str] = None
    func_new_name: Optional[str] = None
    src_module: Optional[str] = None
    src_class: Optional[str] = None
    src_func_name: Optional[str] = None
    can_iterate: bool = True
    func_static_params: Optional[Dict[str, Any]] = None
    func_input_value_arg: Optional[str] = None
    src_config_path: Optional[str] = None
    src_config_entry: Optional[str] = None
    entity_container_name: Optional[str] = None
    return_raw_output: bool = False
    create_shortcut: bool = False

    def attr_for_entity(self, entity: Union[entities.Entity, str]) -> Optional[str]:
        """
        Return the attribute to use for the specified entity.

        Parameters
        ----------
        entity : Union[entities.Entity, str]
            Entity instance or name

        Returns
        -------
        Optional[str]
            Attribute name to use.

        """
        if isinstance(entity, entities.Entity):
            ent_name = entity.__class__.__name__
        else:
            ent_name = entity
        return self.entity_map.get(ent_name)


def create_pivot_func(
    target_func: Callable[[Any], Any],
    pivot_reg: PivotRegistration,
) -> Callable[..., pd.DataFrame]:
    """
    Create function wrapper for pivot function.

    Parameters
    ----------
    target_func: Callable
        The target function to wrap.
    pivot_reg : PivotRegistration
        The pivot function registration object.

    Returns
    -------
    Callable[[Any], pd.DataFrame]
        The original `target_func` wrapped in pre-processing
        and post-processing code.

    """

    @wraps(target_func)
    def pivot_lookup(*args, **kwargs) -> pd.DataFrame:
        """
        Lookup Pivot function from Entity or parameter values.

        Parameters
        ----------
        data: Union[str, List[str], pd.DataFrame]
            Not used if querying the entity value itself

        Returns
        -------
        pd.DataFrame
            DataFrame of Pivot function results.

        """
        # remove and save the join kw, if specified (so it doesn't interfere
        # with other operations and doesn't get sent to the function)
        join_type, left_on, right_on, j_ignore_case = get_join_params(kwargs)

        input_value = _get_input_value(*args, pivot_reg=pivot_reg, parent_kwargs=kwargs)
        _check_valid_settings_for_input(input_value, pivot_reg)

        # If the input_value is not a DF convert it into one and return the DF,
        # the column with the input value(s) plus the param dict that we're going
        # to send to the function. This is going to look like:
        # {"data": input_df, "src_column": input_column}
        input_df, input_column, param_dict = _create_input_df(
            input_value, pivot_reg, parent_kwargs=kwargs
        )

        # Add any static parameters for the function to our params dict
        param_dict.update(pivot_reg.func_static_params or {})

        # Call the target function and collect the results
        if pivot_reg.input_type == "value":
            if not pivot_reg.can_iterate and len(input_df) > 1:
                raise TypeError(
                    "The function does not support multiple input values.",
                    "Try again with a single row/value as input.",
                    "E.g. func(data=df.iloc[N], column=...)",
                )
            result_df = _iterate_func(
                target_func, input_df, input_column, pivot_reg, **kwargs
            )
        else:
            result_df = target_func(**param_dict, **kwargs)  # type: ignore
        merge_key = pivot_reg.func_out_column_name or input_column

        # If requested to join to input
        # and this function is returning a DataFrame
        if join_type and not pivot_reg.return_raw_output:
            left_on = left_on or input_column
            right_on = right_on or merge_key
            return join_result(
                input_df=input_df,
                result_df=result_df,
                how=join_type,
                left_on=left_on,
                right_on=right_on,
                ignore_case=j_ignore_case,
            ).drop(columns="src_row_index", errors="ignore")
        return result_df

    setattr(
        pivot_lookup,
        "pivot_properties",
        attr.asdict(pivot_reg, filter=(lambda _, val: val is not None)),
    )
    return pivot_lookup


def get_join_params(
    func_kwargs: Dict[str, Any]
) -> Tuple[Optional[str], Optional[str], Optional[str], bool]:
    """
    Get join parameters from kwargs.

    Parameters
    ----------
    func_kwargs : Dict[str, Any]
        Keyword arguments from caller

    Returns
    -------
    Tuple[str, str, str, bool]
        join_type, left_on, right_on, join_ignore_case

    """
    # remove and save the join kw, if specified (so it doesn't interfere
    # with other operations and doesn't get sent to the function)
    join_type = func_kwargs.pop("join", None)
    join_ignore_case = func_kwargs.pop("join_ignore_case", None)
    if not join_type:
        return None, None, None, False
    left_on = func_kwargs.pop("left_on", None)
    right_on = func_kwargs.pop("right_on", None)
    if left_on and not right_on:
        warnings.warn(
            "If you are specifying explicit join keys "
            "you must specify 'right_on' parameter with the "
            + "name of the output column to join on. "
            + "Results will joined on index."
        )
    if not left_on:
        col_keys = list(func_kwargs.keys() - {"start", "end", "data"})
        if len(col_keys) == 1:
            # Only one input param so assume this is the src/left
            # join key
            left_on = func_kwargs.get(col_keys[0])

    if right_on and not left_on:
        warnings.warn(
            "Could not infer 'left' join column from source data. "
            + "Please specify 'left_on' parameter with the "
            + "name of the source column to join on. "
            + "Results will joined on index."
        )
    return join_type, left_on, right_on, join_ignore_case


def join_result(
    input_df: pd.DataFrame,
    result_df: pd.DataFrame,
    how: str,
    left_on: str,
    right_on: str,
    ignore_case: bool,
) -> pd.DataFrame:
    """
    Join input and result DFs, optionally ignoring case.

    Parameters
    ----------
    input_df : pd.DataFrame
        Input DF
    result_df : pd.DataFrame
        Result DF
    how : str
        Join type - "inner", "left", "right", "outer"
    left_on : str
        Column from `input_df` to use as join key
    right_on : str
        Column from `result_df` to use as join key
    ignore_case : bool
        If True and input_df column is a string

    Returns
    -------
    pd.DataFrame
        The merged DataFrame

    """
    if not ignore_case or input_df[left_on].dtype.name not in (
        "string",
        "object",
    ):
        # Not requested case-insensitive join OR input column is
        # not a string/object
        return input_df.merge(
            result_df,
            left_on=left_on,
            right_on=right_on,
            how=how,
            suffixes=("_src", "_res"),
        )

    # We need to join case-insensitive
    left_on = f"{left_on}_lc"
    input_df[left_on] = input_df[left_on].str.casefold()
    right_on = f"{right_on}_lc"
    result_df[right_on] = result_df[right_on].astype("string").str.casefold()
    return input_df.merge(
        result_df,
        left_on=left_on,
        right_on=right_on,
        how=how,
        suffixes=("_src", "_res"),
    ).drop(columns=[left_on, right_on])


def _get_entity_attr_or_self(obj, attrib):
    """Return entity attribute or obj if not an entity."""
    if isinstance(obj, entities.Entity):
        return getattr(obj, attrib)
    return obj


def _get_input_value(
    *args, pivot_reg: PivotRegistration, parent_kwargs: Dict[str, Any]
) -> Any:
    """Extract input value from args or kwargs."""
    if args:
        input_value = args[0]
    else:
        # Search possible input arg names
        poss_args = [
            arg
            for arg in [
                pivot_reg.func_df_param_name,
                pivot_reg.func_input_value_arg,
                "value",
                "data",
                "input",
            ]
            if arg
        ]
        for arg_name in poss_args:
            input_value = parent_kwargs.pop(arg_name, None)
            if input_value is not None:
                break
        else:
            raise AttributeError(
                "Required keyword argument not found.",
                f"One of {', '.join(poss_args)} required.",
            )
    if isinstance(input_value, entities.Entity):
        src_entity_attrib = pivot_reg.attr_for_entity(input_value)
        input_value = _get_entity_attr_or_self(input_value, src_entity_attrib)
    return input_value


def _check_valid_settings_for_input(input_value: Any, pivot_reg: PivotRegistration):
    """Check input against settings in `pivot_reg`."""
    # Must have one of these specified
    if not (pivot_reg.func_df_col_param_name or pivot_reg.func_input_value_arg):
        raise ValueError(
            "A value for one of 'func_df_col_param_name' ",
            "or 'func_input_value_arg' must be given",
        )
    # If the function accepts only value type and cannot iterate. Make sure
    # that the input_value is a simple value
    if pivot_reg.input_type == "value":
        if not pivot_reg.func_input_value_arg:
            raise ValueError("No value for pivot func input argument was given")
        if not pivot_reg.can_iterate and (
            isinstance(input_value, pd.DataFrame)
            or (
                # pylint: disable=isinstance-second-argument-not-valid-type
                isinstance(input_value, pd.DataFrame)
                and not isinstance(input_value, str)
                # pylint: enable=isinstance-second-argument-not-valid-type
            )
        ):
            raise ValueError(
                f"This function does not accept inputs of {type(input_value)}"
            )


def _arg_to_dframe(arg_val, col_name: str = "param_value"):
    """
    Convert a scalar or Iterable value to a DataFrame.

    Parameters
    ----------
    arg_val: Any
        The value to be converted
    col_name: Optional[str]
        The name to assign to the DataFrame column

    Returns
    -------
    pd.DataFrame
        The resulting DataFrame

    Notes
    -----
    If `arg_val` is already a DataFrame it is returned as is.

    """
    if isinstance(arg_val, pd.DataFrame):
        return arg_val
    if isinstance(arg_val, str) or not isinstance(arg_val, abc.Iterable):
        return pd.DataFrame([arg_val], columns=[col_name])
    return pd.DataFrame(arg_val, columns=[col_name])


def _create_input_df(input_value, pivot_reg, parent_kwargs):
    """Create input_df and params from input."""
    # If input_value type is not already a dataframe, convert it.
    # If the DF column param is specified, use that or fall back
    # to using the function input value arg.
    input_column = pivot_reg.func_df_col_param_name or pivot_reg.func_input_value_arg
    # If input_value is already a DF, this call just returns the original DF
    input_df = _arg_to_dframe(input_value, input_column)  # type: ignore

    if isinstance(input_value, pd.DataFrame):
        # If the original input_value is a DataFrame
        # try to find the column name specification in kwargs
        for col_param in (
            pivot_reg.func_df_col_param_name,
            pivot_reg.func_input_value_arg,
            *_DF_SRC_COL_PARAM_NAMES,
        ):
            if col_param in parent_kwargs and parent_kwargs[col_param] in input_df:
                input_column = parent_kwargs.pop(col_param)
                break
        else:
            col_arg = parent_kwargs.pop(col_param, None)
            col_arg = f" '{col_arg}' " if col_arg else ""
            col_mssg = f"The column name {col_arg} given for the 'column' parameter"
            raise KeyError(
                f"'{col_mssg}' is not in the input dataframe",
                "Please specify the column when calling the function."
                "You can use one of these parameter names for this:",
                _DF_SRC_COL_PARAM_NAMES,
            )
    # we want to get rid of data=xyz parameters from kwargs, since we're adding them
    # below
    parent_kwargs.pop("data", None)
    parent_kwargs.pop(pivot_reg.func_df_param_name, None)

    if input_column not in input_df:
        raise KeyError(f"'{input_column}' is not in the input dataframe")
    if input_column:
        param_dict = {
            pivot_reg.func_df_param_name: input_df,
            pivot_reg.func_df_col_param_name: input_column,
        }
    else:
        # If no column was specified, the user will have to specify
        # this in the call to the method - we just add the DF parameter
        param_dict = {pivot_reg.func_df_param_name: input_df}
    return input_df, input_column, param_dict


def _iterate_func(target_func, input_df, input_column, pivot_reg, **kwargs):
    """Call `target_func` function with values of each row in `input_df`."""
    results = []
    # Add any static parameters to all_rows_kwargs
    all_rows_kwargs = kwargs.copy()
    all_rows_kwargs.update((pivot_reg.func_static_params or {}))
    res_key_col_name = pivot_reg.func_out_column_name or pivot_reg.func_input_value_arg

    for row_index, row in enumerate(input_df[[input_column]].itertuples(index=False)):
        # Get rid of any conflicting arguments from kwargs
        func_kwargs = all_rows_kwargs.copy()
        func_kwargs.pop(pivot_reg.func_input_value_arg, None)
        # Create a param dictionary with the value parameter for this row
        param_dict = {pivot_reg.func_input_value_arg: row[0]}
        # run the function
        result = target_func(**param_dict, **all_rows_kwargs)

        # Process the output, if it is a DataFrame
        if not pivot_reg.return_raw_output and not isinstance(result, pd.DataFrame):
            col_value = next(iter(row._asdict().values()))
            if isinstance(result, dict):
                # if result is a dict - make that into a row.
                result = pd.DataFrame(pd.Series(result)).T
                result[res_key_col_name] = col_value
            else:
                # just make the result into a string and use that as a single col
                result = pd.DataFrame(
                    [[col_value, str(result)]], columns=[res_key_col_name, "result"]
                )
            result["src_row_index"] = row_index
        results.append(result)
    if pivot_reg.return_raw_output:
        if len(results) == 1:
            return results[0]
        return results
    return pd.concat(results, ignore_index=True)


# _PARENT_SELF = "parent_self"


# def query_cont_member_wrap(func: Callable[[Any], Any]) -> Callable[[Any], Any]:
#     """
#     Wrap a func to work as instance method in a QueryContainer.

#     Parameters
#     ----------
#     func : Callable[[Any], Any]
#         Function to wrap as method

#     Returns
#     -------
#     Callable[[Any], Any]
#         Wrapped function

#     Notes
#     -----
#     This is designed to be used inside a `QueryContainer`. The wrapped
#     function checks to see if its arg[0] is a QueryContainer - meaning
#     it has been called as an instance function of that class.
#     If so, and the parent class has a _parent_self attribute, it will
#     replace the original arg[0] (the self of QueryContainer) with
#     the self of the containing class (_parent_self).
#     It relies containing class setting `_parent_self` as an attribute
#     in any QueryContainer attributes that it has. The msticpy Entity
#     class does this.

#     If these conditions don't apply it simply passed through the call
#     to the original function.

#     See Also
#     --------
#     QueryContainer
#     Entity

#     """

#     @wraps(func)
#     def _wrapped_member(*args, **kwargs):
#         if (
#             args
#             and args[0].__class__.__name__ == "QueryContainer"
#             and hasattr(args[0], _PARENT_SELF)
#         ):
#             parent_self = getattr(args[0], _PARENT_SELF)
#             return func(parent_self, *args[1:], **kwargs)
#         return func(*args, **kwargs)

#     return _wrapped_member



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Kusto Driver subclass."""
from typing import Any, Dict, Optional, Union

import pandas as pd

from ...common.exceptions import MsticpyParameterError, MsticpyUserConfigError
from ...common.provider_settings import get_provider_settings, ProviderArgs
from ...common.utility import export
from ..query_defns import DataEnvironment
from .kql_driver import KqlDriver, QuerySource

from ..._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"

_KCS_CODE = "code;"
_KCS_APP = "tenant='{tenant}';clientid='{clientid}';clientsecret='{clientsecret}';"
_KCS_TEMPLATE = "azure_data-Explorer://{auth}cluster='{cluster}';database='{database}'"

KustoClusterSettings = Dict[str, Dict[str, Union[str, ProviderArgs]]]


@export
class KustoDriver(KqlDriver):
    """Kusto Driver class to execute kql queries for Azure Data Explorer."""

    def __init__(self, connection_str: str = None, **kwargs):
        """
        Instantiate KustoDriver.

        Parameters
        ----------
        connection_str : str, optional
            Connection string

        Other Parameters
        ----------------
        debug : bool
            print out additional diagnostic information.

        """
        super().__init__(connection_str=connection_str, **kwargs)
        self.environment = kwargs.get("data_environment", DataEnvironment.Kusto)
        self._connected = True
        self._kusto_settings: KustoClusterSettings = _get_kusto_settings()

    def connect(self, connection_str: Optional[str] = None, **kwargs):
        """
        Connect to data source.

        Parameters
        ----------
        connection_str : str
            Connect to a data source

        Other Parameters
        ----------------
        cluster : str, optional
            Short name or URI of cluster to connect to.
        database : str, optional
            Name of database to connect to.
        kqlmagic_args : str, optional
            Additional string of parameters to be passed to KqlMagic
        mp_az_auth : Union[bool, str, list, None], optional
            Optional parameter directing KqlMagic to use MSTICPy Azure authentication.
            Values can be:
            True or "default": use the settings in msticpyconfig.yaml 'Azure' section
            str: single auth method name ('msi', 'cli', 'env' or 'interactive')
            List[str]: list of acceptable auth methods from ('msi', 'cli',
            'env' or 'interactive')

        """
        self.current_connection = self._get_connection_string(
            connection_str=connection_str, **kwargs
        )
        kwargs.pop("cluster", None)
        kwargs.pop("database", None)
        super().connect(connection_str=self.current_connection, **kwargs)

    def query(
        self, query: str, query_source: QuerySource = None, **kwargs
    ) -> Union[pd.DataFrame, Any]:
        """
        Execute query string and return DataFrame of results.

        Parameters
        ----------
        query : str
            The query to execute
        query_source : QuerySource
            The query definition object

        Other Parameters
        ----------------
        cluster : str, Optional
            Supply or override the Kusto cluster name
        database : str, Optional
            Supply or override the Kusto database name
        data_source : str, Optional
            alias for `db`
        connection_str : str, Optional


        Returns
        -------
        Union[pd.DataFrame, results.ResultSet]
            A DataFrame (if successfull) or
            the underlying provider result if an error.

        """
        new_connection = self._get_connection_string(
            query_source=query_source, **kwargs
        )
        if new_connection:
            self.current_connection = new_connection
        data, result = self.query_with_results(query)
        return data if data is not None else result

    def _get_connection_string(self, query_source: QuerySource = None, **kwargs):
        """Create a connection string from arguments and configuration."""
        # If the connection string is supplied as a parameter, use that
        cluster = database = None
        new_connection = kwargs.get("connection_str")
        if not new_connection:
            # try to get cluster and db from kwargs or query_source metadata
            cluster = self._lookup_cluster(kwargs.get("cluster", ""))
            database = kwargs.get("database")
            if cluster and database:
                new_connection = self._create_connection(
                    cluster=cluster, database=database
                )
        if not new_connection and query_source:
            # try to get cluster and db from query_source metadata
            cluster = cluster or query_source.metadata.get("cluster")
            data_families = query_source.metadata.get("data_families")
            if (
                not isinstance(data_families, list) or len(data_families) == 0
            ) and not self.current_connection:
                # call create connection so that we throw an informative error
                self._create_connection(cluster=cluster, database=database)
            if "." in data_families[0]:  # type: ignore
                _, qry_db = data_families[0].split(".", maxsplit=1)  # type: ignore
            else:
                # Not expected but we can still use a DB value with no dot
                qry_db = data_families[0]  # type: ignore
            database = database or qry_db
            new_connection = self._create_connection(cluster=cluster, database=database)
        return new_connection

    def _create_connection(self, cluster, database):
        """Create the connection string, checking parameters."""
        if not cluster or not database:
            if cluster:
                err_mssg = "database name"
            elif database:
                err_mssg = "cluster uri"
            else:
                err_mssg = "cluster uri and database name"
            raise MsticpyParameterError(
                f"Could not determine the {err_mssg} for the query.",
                "Please update the query with the correct values or specify",
                "explicitly with the 'cluster' and 'database' parameters to",
                "this function.",
                "In the query template these values are specified in the metadata:",
                "cluster: cluster_uri",
                "data_families: [ClusterAlias.database]",
                title="Missing cluster or database names.",
                parameter=err_mssg,
            )
        cluster_key = cluster.casefold()
        if cluster_key not in self._kusto_settings:
            raise MsticpyUserConfigError(
                f"The cluster {cluster} was not found in the configuration.",
                "You must have an entry for the cluster in the 'DataProviders section",
                "of your msticyconfig.yaml",
                "Expected format:",
                "Kusto[-instance_name]:",
                "  args:",
                "    Cluster: cluster_uri",
                "    Integrated: True",
                "or",
                "Kusto[-instance_name]:",
                "  args:",
                "    Cluster: cluster_uri",
                "    TenantId: tenant_uuid",
                "    ClientId: tenant_uuid",
                "    ClientSecret: (string|KeyVault|EnvironmentVar:)",
                title="Unknown cluster.",
            )
        return self._format_connection_str(cluster, database)

    def _format_connection_str(self, cluster: str, database: str) -> Optional[str]:
        """Return connection string with client secret added."""
        fmt_items = self._kusto_settings.get(cluster.casefold())
        if not fmt_items:
            return None
        fmt_items["database"] = database
        if fmt_items.get("integrated_auth"):
            auth_string = _KCS_CODE
        else:
            # Note, we don't add the secret until required at runtime to prevent
            # it hanging around in memory as much as possible.
            fmt_items["clientsecret"] = fmt_items["args"].get("ClientSecret")  # type: ignore
            auth_string = _KCS_APP.format(**fmt_items)
        return _KCS_TEMPLATE.format(auth=auth_string, **fmt_items)

    def _lookup_cluster(self, cluster: str):
        """Return cluster URI from config if cluster name is passed."""
        if cluster.strip().casefold().startswith("https://"):
            return cluster
        for cluster_key, kusto_config in self._kusto_settings.items():
            if cluster_key.startswith(f"https://{cluster.casefold()}."):
                return kusto_config["cluster"]
        return None


def _get_kusto_settings() -> KustoClusterSettings:
    kusto_settings: KustoClusterSettings = {}
    for prov_name, settings in get_provider_settings("DataProviders").items():
        if not prov_name.startswith("Kusto"):
            continue
        instance = "Kusto"
        if "-" in prov_name:
            _, instance = prov_name.split("-", maxsplit=1)

        cluster = settings.args.get("Cluster")
        if not cluster:
            raise MsticpyUserConfigError(
                "Mandatory 'Cluster' setting is missing in msticpyconfig.",
                f"the Kusto entry with the missing setting is '{prov_name}'",
                title=f"No Cluster value for {prov_name}",
            )
        kusto_settings[cluster.casefold()] = {
            "tenant": settings.args.get("TenantId"),  # type: ignore
            "integrated_auth": settings.args.get("IntegratedAuth"),  # type: ignore
            "clientid": settings.args.get("ClientId"),  # type: ignore
            "args": settings.args,
            "cluster": cluster,
            "alias": instance,
        }
    return kusto_settings



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Account Entity class."""
from typing import Any, Mapping, Optional

from ..._version import VERSION
from ...common.utility import export
from .entity import Entity
from .host import Host

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=invalid-name, too-many-instance-attributes


@export
class Account(Entity):
    """
    Account Entity class.

    Attributes
    ----------
    Name : str
        Account Name
    NTDomain : str
        Account NTDomain
    UPNSuffix : str
        Account UPNSuffix
    Host : Host
        Account Host
    LogonId : str
        Account LogonId (deprecated)
    Sid : str
        Account Sid
    AadTenantId : str
        Account AadTenantId
    AadUserId : str
        Account AadUserId
    PUID : str
        Account PUID
    IsDomainJoined : bool
        Account IsDomainJoined
    DisplayName : str
        Account DisplayName
    ObjectGuid : str
        The object ID of the user account

    """

    ID_PROPERTIES = ["QualifiedName", "Sid", "AadUserId", "PUID", "ObjectGuid"]

    def __init__(
        self,
        src_entity: Mapping[str, Any] = None,
        src_event: Mapping[str, Any] = None,
        role: str = "subject",
        **kwargs,
    ):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing Account entity or
            other mapping object that implements entity properties.
            (the default is None)
        src_event : Mapping[str, Any], optional
            Create entity from event properties
            (the default is None)
        role : str, optional
            'subject' or 'target' - only relevant if the entity
            is being constructed from an event.
            (the default is 'subject')

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.

        """
        self.Name: Optional[str] = None
        self.NTDomain: Optional[str] = None
        self.UPNSuffix: Optional[str] = None
        self.Host: Optional[Host] = None
        self.LogonId: Optional[str] = None
        self.Sid: Optional[str] = None
        self.AadTenantId: Optional[str] = None
        self.AadUserId: Optional[str] = None
        self.PUID: Optional[str] = None
        self.IsDomainJoined: bool = False
        self.DisplayName: Optional[str] = None
        self.ObjectGuid: Optional[str] = None

        # pylint: disable=locally-disabled, line-too-long
        super().__init__(src_entity=src_entity, **kwargs)
        if src_event is not None:
            self._create_from_event(src_event, role)

    # pylint: enable=locally-disabled, line-too-long

    @property
    def description_str(self) -> str:
        """Return Entity Description."""
        return self.qualified_name

    @property
    def name_str(self) -> str:
        """Return Entity Name."""
        return self.Name or self.DisplayName or "Unknown Account"

    @property
    def qualified_name(self) -> str:
        """Windows qualified account name."""
        if "Name" not in self:
            return ""
        name = self["Name"]
        if "NTDomain" in self and self.NTDomain:
            return f"{self.NTDomain}\\{name}"
        if "UPNSuffix" in self and self.UPNSuffix:
            return f"{name}@{self.UPNSuffix}"
        if "Host" in self and self.Host:
            return f"{self.Host.HostName}\\{name}"
        return name

    def _create_from_event(self, src_event, role):
        if "Name" in src_event:
            self.Name = src_event["Name"]
        if role == "subject" and "SubjectUserName" in src_event:
            self.Name = src_event["SubjectUserName"]
            self.NTDomain = (
                src_event["SubjectUserDomain"]
                if "SubjectUserDomain" in src_event
                else None
            )
            self.Sid = (
                src_event["SubjectUserSid"] if "SubjectUserSid" in src_event else None
            )
            self.LogonId = (
                src_event["SubjectLogonId"] if "SubjectLogonId" in src_event else None
            )
        if role == "target" and "TargetUserName" in src_event:
            self.Name = src_event["TargetUserName"]
            self.NTDomain = (
                src_event["TargetUserDomain"]
                if "TargetUserDomain" in src_event
                else None
            )
            self.Sid = (
                src_event["TargetUserSid"] if "TargetUserSid" in src_event else None
            )
            self.LogonId = (
                src_event["TargetLogonId"] if "TargetLogonId" in src_event else None
            )

        self.AadTenantId = (
            src_event["AadTenantId"] if "AadTenantId" in src_event else None
        )
        self.Sid = src_event["Sid"] if "Sid" in src_event else None
        self.NTDomain = src_event["NtDomain"] if "NtDomain" in src_event else None
        self.AadUserId = src_event["AadUserId"] if "AadUserId" in src_event else None
        self.PUID = src_event["PUID"] if "PUID" in src_event else None
        if "DisplayName" in src_event:
            self.DisplayName = src_event["DisplayName"]
        elif "AccountName" in src_event:
            self.DisplayName = src_event["AccountName"]
        else:
            self.DisplayName = None

        if "UPNSuffix" in src_event:
            self.UPNSuffix = src_event["UPNSuffix"]
        elif "UpnSuffix" in src_event:
            self.UPNSuffix = src_event["UpnSuffix"]
        else:
            self.UPNSuffix = None

    _entity_schema = {
        # Name (type System.String)
        "Name": None,
        # NTDomain (type System.String)
        "NTDomain": None,
        # UPNSuffix (type System.String)
        "UPNSuffix": None,
        # Host (type Microsoft.Azure.Security.Detection
        # .AlertContracts.V3.Entities.Host)
        "Host": "Host",
        # LogonId (type System.String)
        "LogonId": None,
        # Sid (type System.String)
        "Sid": None,
        # AadTenantId (type System.Nullable`1[System.Guid])
        "AadTenantId": None,
        # AadUserId (type System.Nullable`1[System.Guid])
        "AadUserId": None,
        # PUID (type System.Nullable`1[System.Guid])
        "PUID": None,
        # IsDomainJoined (type System.Nullable`1[System.Boolean])
        "IsDomainJoined": None,
        # DisplayName (type System.String)
        "DisplayName": None,
        "ObjectGuid": None,
        "TimeGenerated": None,
        "StartTime": None,
        "EndTime": None,
    }



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""QueryProvider Query Browser."""
import textwrap
from typing import Generator, Any

from IPython.core.display import HTML

from ..._version import VERSION
from ...nbtools.nbwidgets import SelectItem

__version__ = VERSION
__author__ = "Ian Hellen"


def browse_queries(query_provider: Any, **kwargs) -> SelectItem:
    """
    Return QueryProvider query browser.

    Parameters
    ----------
    query_provider : QueryProvider
        Initialized query provider.

    Other Parameters
    ----------------
    kwargs :
        passed to SelectItem constuctor.

    Returns
    -------
    SelectItem
        SelectItem browser for TI Data.

    """
    if "height" not in kwargs:
        kwargs["height"] = "300px"
    disp_func = _query_display_func(query_provider)
    opts = query_provider.list_queries()
    return SelectItem(item_list=opts, action=disp_func, **kwargs)


def _format_query_doc(query_doc) -> Generator[str, None, None]:
    """Format query docstring as HTML."""
    qdoc_lines = query_doc.split("\n")
    yield f"<h3>{qdoc_lines[0]}</h3>"
    for line in qdoc_lines[1:]:
        if line.strip() == "Parameters":
            yield f"<p><b>{line}</b></p>"
        elif line.strip().startswith("---"):
            continue
        else:
            indent = len(line) - len(line.strip())
            if indent:
                indent *= 10
                yield f"<div style='margin-left: {indent}px'>{line}</div>"
            else:
                yield f"<div>{line}</div>"


def _format_query(query_text):
    """Format query for display."""
    q_lines = []
    for line in query_text.split("|"):
        if len(line) > 80:
            line = "\n  ".join(textwrap.wrap(line))
        q_lines.append(line)
    q_text = "\n|".join(q_lines)
    return f"<p><b>Query</b></p><pre>{q_text}</pre>"


def _get_query_sample(query):
    """Format query sample."""
    return f"""
        <p><b>Example</b></p>
        <p>{{QueryProvider}}[.QueryPath].QueryName(params...)</p>
        <pre>qry_prov.{query}(start=start, end=end, hostname=host)</pre>
    """


def _query_display_func(query_provider):
    """Closure returning function to display query."""

    def get_query_doc(query):
        """Return doc string and query as HTML."""
        query_func = getattr(query_provider, query)
        q_help = "".join(_format_query_doc(query_func.__doc__))
        q_text = query_provider.get_query(query).strip()
        q_sample = _get_query_sample(query)
        q_text = _format_query(q_text)
        return HTML(f"{q_help}<br>{q_text}<br>{q_sample}")

    return get_query_doc



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""File Entity class."""
from typing import Any, List, Mapping, Optional

from ..._version import VERSION
from ...common.utility import export
from .entity import Entity
from .entity_enums import Algorithm, OSFamily
from .file_hash import FileHash
from .host import Host

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=invalid-name, too-many-instance-attributes


@export
class File(Entity):
    """
    File Entity class.

    Attributes
    ----------
    FullPath : str
        File FullPath
    Directory : str
        File Directory
    Name : str
        File Name
    Md5 : str
        File Md5
    Host : str
        File Host
    Sha1 : str
        File Sha1
    Sha256 : str
        File Sha256
    Sha256Ac : str
        File Sha256Ac
    FileHashes : List[FileHash]
        File FileHashes

    """

    ID_PROPERTIES = ["FullPath", "Sha1", "Sha256", "Sha256ac", "Md5"]

    def __init__(
        self,
        src_entity: Mapping[str, Any] = None,
        src_event: Mapping[str, Any] = None,
        role: str = "new",
        **kwargs,
    ):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing entity or
            other mapping object that implements entity properties.
            (the default is None)
        src_event : Mapping[str, Any], optional
            Create entity from event properties
            (the default is None)
        role : str, optional
            'new' or 'parent' - only relevant if the entity
            is being constructed from an event.
            (the default is 'new')

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.

        """
        self.FullPath: Optional[str] = None
        self.Directory: Optional[str] = None
        self.Name: Optional[str] = None
        self.Md5: Optional[str] = None
        self.Host: Optional[Host] = None
        self.Sha1: Optional[str] = None
        self.Sha256: Optional[str] = None
        self.Sha256Ac: Optional[str] = None
        self.FileHashes: List[FileHash] = []
        self.PathSeparator: Optional[str] = "\\"
        self.OSFamily = OSFamily.Windows
        super().__init__(src_entity=src_entity, **kwargs)
        if src_event is not None:
            self._create_from_event(src_event, role)

        if not self.FullPath:
            file_name = self.Name
            directory = self.Directory or ""
            sep = self.path_separator if directory else ""
            self.FullPath = f"{directory}{sep}{file_name}"

    @property
    def path_separator(self):
        """Return the path separator used by the file."""
        if (
            self.Directory and "/" in self.Directory
        ) or self.OSFamily != OSFamily.Windows:
            return "/"
        return "\\"

    @property
    def description_str(self) -> str:
        """Return Entity Description."""
        return self.FullPath or self.__class__.__name__

    @property
    def name_str(self) -> str:
        """Return Entity Name."""
        return self.Name or self.FullPath or self.__class__.__name__

    def _create_from_event(self, src_event, role):
        if role == "new" and "NewProcessName" in src_event:
            self._add_paths(src_event["NewProcessName"])
        elif role == "parent" and "ParentProcessName" in src_event:
            self._add_paths(src_event["ParentProcessName"])
        elif "Directory" in src_event and "FileName" in src_event:
            self._add_paths(src_event["Directory"], file_name=src_event["FileName"])

    _entity_schema = {
        # FullPath (type System.String)
        "FullPath": None,
        # Directory (type System.String)
        "Directory": None,
        # Name (type System.String)
        "Name": None,
        # Md5 (type System.String)
        "Md5": None,
        # Host (type Microsoft.Azure.Security.Detection
        # .AlertContracts.V3.Entities.Host)
        "Host": "Host",
        # Sha1 (type System.String)
        "Sha1": None,
        # Sha256 (type System.String)
        "Sha256": None,
        # Sha256Ac (type System.String)
        "Sha256Ac": None,
        "FileHashes": (list, "FileHash"),
        "TimeGenerated": None,
        "StartTime": None,
        "EndTime": None,
    }

    def _add_paths(self, full_path, file_name=None):
        if "/" in full_path:
            self.PathSeparator = "/"
            self.OSFamily = OSFamily.Linux
        else:
            self.PathSeparator = "\\"
            self.OSFamily = OSFamily.Windows

        if file_name:
            self.Name = file_name
            self.Directory = full_path
            self.FullPath = self.Directory + self.PathSeparator + self.Name
        else:
            self.FullPath = full_path
            self.Name = full_path.split(self.PathSeparator)[-1]
            self.Directory = full_path.split(self.PathSeparator)[:-1]

    @property
    def file_hash(self) -> Optional[str]:
        """
        Return the first defined file hash.

        Returns
        -------
        Optional[str]
            Returns first-defined file hash in order of
            SHA256, SHA1, MD5, SHA256AC (authenticode)

        """
        if self.FileHashes:
            alg_order = {
                Algorithm.SHA256: 0,
                Algorithm.SHA1: 1,
                Algorithm.MD5: 2,
                Algorithm.SHA256AC: 3,
            }

            return next(
                iter(
                    sorted(
                        self.FileHashes,
                        key=lambda f_hash: alg_order.get(f_hash.Algorithm, 10),
                    )
                )
            ).Value

        return self.Sha256 or self.Sha1 or self.Md5 or self.Sha256Ac



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Incident Entity class."""
from datetime import datetime
from typing import Any, List, Mapping, Optional, Dict
import pandas as pd

from ..._version import VERSION
from ...common.utility import export
from ..entities.entity import Entity, camelcase_property_names
from ..entities.alert import Alert

__version__ = VERSION
__author__ = "Pete Bryan"


# pylint: disable=invalid-name, too-many-instance-attributes
@export
class Incident(Entity):
    """Incident Entity class."""

    ID_PROPERTIES = ["IncidentID"]

    def __init__(
        self,
        src_entity: Mapping[str, Any] = None,
        src_event: Mapping[str, Any] = None,
        src_event_type: str = "Sentinel",
        **kwargs,
    ):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing entity or
            other mapping object that implements entity properties.
            (the default is None)
        src_event : Mapping[str, Any], optional
            Create entity from event properties
            (the default is None)
        src_event_type: str
            The type of src_event, by default "Sentinel".

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.

        """
        self.TimeGenerated: Optional[datetime] = None
        self.DisplayName: Optional[str] = None
        self.IncidentID: Optional[str] = None
        self.Severity: Optional[str] = None
        self.Status: Optional[str] = None
        self.Owner: Optional[Dict] = None
        self.Classification: Optional[str] = None
        self.Labels: Optional[List] = None
        self.Alerts: Optional[List] = None
        self.Entities: Optional[List] = None

        super().__init__(src_entity=src_entity, **kwargs)

        if src_entity:
            self._create_from_sent_event(src_entity)

        if isinstance(src_event, pd.Series) and src_event_type == "Sentinel":
            self._create_from_sent_event(src_event)

    @property
    def description_str(self) -> str:
        """Return Entity Description."""
        return f"{self.TimeGenerated} - {self.DisplayName} - {self.IncidentID}"

    @property
    def name_str(self) -> str:
        """Return Entity Description."""
        return f"Incident: {self.DisplayName}"

    def _create_from_sent_event(self, src_event):
        self.TimeGenerated = src_event["properties.createdTimeUtc"]
        self.DisplayName = src_event["properties.title"]
        self.IncidentID = src_event["id"].split("/")[-1]
        self.Severity = src_event["properties.severity"]
        self.Status = src_event["properties.status"]
        if "properties.classification" in src_event:
            self.Classification = src_event["properties.classification"]
        self.Labels = src_event["properties.labels"]
        owner = {
            "Name": src_event["properties.owner.assignedTo"],
            "Email": src_event["properties.owner.email"],
            "UPN": src_event["properties.owner.userPrincipalName"],
            "ID": src_event["properties.owner.objectId"],
        }
        self.Owner = owner
        additionaldata = {
            "Alert Count": src_event["properties.additionalData.alertsCount"],
            "Bookmarks Count": src_event["properties.additionalData.bookmarksCount"],
            "Comments Count": src_event["properties.additionalData.commentsCount"],
            "Products": src_event["properties.additionalData.alertProductNames"],
            "Tactics": src_event["properties.additionalData.tactics"],
            "Related Rule IDs": src_event["properties.relatedAnalyticRuleIds"],
            "Incident Number": src_event["properties.incidentNumber"],
        }
        self.AdditionalData = additionaldata
        self.StartTime = src_event["properties.firstActivityTimeUtc"]
        self.EndTime = src_event["properties.lastActivityTimeUtc"]
        if "Alerts" in src_event:
            self._add_alerts(src_event)

        if "Entities" in src_event:
            self.Entities = _create_entities(src_event["Entities"])

    def _add_alerts(self, src_event):
        """Add alerts to incident."""
        if src_event["Alerts"] and isinstance(src_event["Alerts"], list):
            new_alerts = [Alert(src_entity=alrt) for alrt in src_event["Alerts"]]
            self.Alerts = new_alerts

    _entity_schema = {
        # Time the Incident was Generated
        "TimeGenerated": None,
        # Incident Name
        "DisplayName": None,
        # Some unique identifier
        "IncidentID": None,
        # Incident severity
        "Severity": None,
        # Current status of the incident i.e. Open, Closed, etc
        "Status": None,
        # Object containing details of incident owner, could be email, display name, etc
        "Owner": None,
        # Post triage setting i.e. FP, TP, etc
        "Classification": None,
        # List of labels applied to an icident
        "Labels": None,
        # Assocaited alert GUIDs
        "Alerts": None,
        # List of associated entities
        "Entities": None,
        # Dynamic bag of other items
        "AdditionalData ": None,
        "StartTime": None,
        "EndTime": None,
    }


def _create_entities(entities):
    """Create incident entities from API returned dicts."""
    new_ents = []
    for ent in entities:
        if isinstance(ent, tuple):
            ent_details = ent[1]
            ent_type = ent[0]
        elif isinstance(ent, dict):
            ent_details = ent
            ent_type = ent["Type"]
        else:
            ent_details = ent
            ent_type = "unknown"
        new_ent = camelcase_property_names(ent_details)
        ent_obj = Entity.ENTITY_NAME_MAP[ent_type.lower()](src_event=new_ent)
        new_ents.append(ent_obj)
    return new_ents



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Submission mail Entity class."""
from typing import Any, Mapping, Optional

from ..._version import VERSION
from ...common.utility import export
from .entity import Entity

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=invalid-name


@export
class SubmissionMail(Entity):
    """
    SubmissionMail Entity class.

    Attributes
    ----------
    SubmissionId : str
        SubmissionId of the submission mail
    SubmissionDate : datetime
        SubmissionDate of the submission mail
    Submitter : str
        Submitter email of the submission mail
    NetworkMessageId : str
        NetworkMessageId of the submission mail
    Timestamp : datetime
        The Time stamp when the message is received
    Recipient : str
        Recipient of the submission mail
    Sender : str
        Sender of the submission mail
    SenderIp : str
        SenderIp of the submission mail
    Subject : str
        Subject of the submission mail
    ReportType : str
        ReportType of the submission mail

    """

    ID_PROPERTIES = ["NetworkMessageId", "Recipient"]

    def __init__(self, src_entity: Mapping[str, Any] = None, **kwargs):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing entity or
            other mapping object that implements entity properties.
            (the default is None)

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.

        """
        self.SubmissionId: Optional[str] = None
        self.SubmissionDate: Any = None
        self.Submitter: Optional[str] = None
        self.NetworkMessageId: Optional[str] = None
        self.Timestamp: Any = None
        self.Recipient: Optional[str] = None
        self.Sender: Optional[str] = None
        self.SenderIp: Optional[str] = None
        self.Subject: Optional[str] = None
        self.ReportType: Optional[str] = None

        super().__init__(src_entity=src_entity, **kwargs)

    @property
    def description_str(self):
        """Return Entity Description."""
        return self.SubmissionId or self.__class__.__name__

    @property
    def name_str(self) -> str:
        """Return Entity Name."""
        return self.__class__.__name__

    _entity_schema = {
        "SubmissionId": None,
        "SubmissionDate": None,
        "Submitter": None,
        "NetworkMessageId": None,
        "Timestamp": None,
        "Recipient": None,
        "Sender": None,
        "SenderIp": None,
        "Subject": None,
        "ReportType": None,
        "TimeGenerated": None,
        "StartTime": None,
        "EndTime": None,
    }



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Parameter extractor helper functions for use with IPython/Juptyer queries."""
from typing import List, Dict, Mapping, Tuple, Any

from .query_store import QuerySource
from ..common.utility import export
from .query_defns import QueryParamProvider
from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


@export
def extract_query_params(
    query_source: QuerySource, *args, **kwargs
) -> Tuple[Dict[str, Any], List[str]]:
    """
    Get the parameters needed for the query.

    Parameters
    ----------
    query_source : QuerySource
        Query source
    args : Tuple[QueryParamProvider]
        objects that implement QueryParamProvider
        (from which query parameters can be extracted).
    kwargs : Dict[str, Any]
        custom parameter list to populate queries
        (override default values and values extracted
        from QueryParamProviders).

    Returns
    -------
    Tuple[Dict[str, Any], List[str]]
        Dictionary of parameter names and values to be used
        in the query.
        List of any missing parameters

    """
    # get the parameters for this query
    all_params = query_source.params

    # required_params are those that don't have defaults set in the query
    # template. Build a dictionary to hold the values. This will contain
    # at least the required params plus any that are extracted from args and
    # kwargs and have been added dynamically.
    req_param_names = query_source.required_params.keys()
    req_params: Dict[str, Any] = {param: None for param in req_param_names}

    # try to retrieve any parameters as attributes of the args objects
    _get_object_params(args, all_params, req_params)

    # If any kwargs parameters have been supplied, add these.
    # These any parameters obtained from _get_object_params
    if kwargs:
        resolved_params = query_source.resolve_param_aliases(kwargs)
        req_params.update(
            {key: arg for key, arg in resolved_params.items() if key in all_params}
        )

    # Get the names of any params that were required but we didn't
    # find a value for
    missing_params = [
        p_name for p_name, p_value in req_params.items() if p_value is None
    ]
    return req_params, missing_params


def _get_object_params(
    args: Tuple[Any, ...], params: Mapping[str, Any], req_params: Dict[str, Any]
):
    """
    Get params from attributes of arg objects.

    Parameters
    ----------
    args : Tuple[Any]
        Args list from calling function
    params : Mapping[str, Dict[str, Any]]
        The list of parameter names to look for
    req_params : Dict[str, str]
        Dictionary of required parameters

    """
    remaining_params = list(params.keys())
    for arg_object in args:
        if isinstance(arg_object, (str, int, float, bool)):
            # ignore some common primitive types
            continue
        for param in remaining_params:
            if isinstance(arg_object, QueryParamProvider):
                if param in arg_object.query_params:
                    req_params[param] = arg_object.query_params[param]
            elif isinstance(arg_object, dict) and param in arg_object:
                req_params[param] = arg_object.get(param, None)
            elif hasattr(arg_object, param):
                req_params[param] = getattr(arg_object, param)
        remaining_params = [
            p_name for p_name, p_value in req_params.items() if p_value is not None
        ]



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Pandas accessor class for timeline functions."""
from typing import Iterable, Optional, Union
import pandas as pd

from bokeh.models import LayoutDOM

from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"

# pylint: disable=unused-import, import-outside-toplevel, cyclic-import


@pd.api.extensions.register_dataframe_accessor("mp_timeline")
class TimeLineAccessor:
    """Pandas api extension for Timeline."""

    def __init__(self, pandas_obj):
        """Instantiate pandas extension class."""
        from .timeline import display_timeline, display_timeline_values
        from .timeline_duration import display_timeline_duration

        self._display_timeline = display_timeline
        self._display_timeline_values = display_timeline_values
        self._display_timeline_duration = display_timeline_duration
        self._df = pandas_obj

    def plot(self, **kwargs) -> LayoutDOM:
        """
        Display a timeline of events.

        Parameters
        ----------
        time_column : str, optional
            Name of the timestamp column
            (the default is 'TimeGenerated')
        source_columns : list, optional
            List of default source columns to use in tooltips
            (the default is None)

        Other Parameters
        ----------------
        title : str, optional
            Title to display (the default is None)
        alert : SecurityAlert, optional
            Add a reference line/label using the alert time (the default is None)
        ref_event : Any, optional
            Add a reference line/label using the alert time (the default is None)
        ref_time : datetime, optional
            Add a reference line/label using `ref_time` (the default is None)
        group_by : str
            The column to group timelines on.
        legend: str, optional
            "left", "right", "inline" or "none"
            (the default is to show a legend when plotting multiple series
            and not to show one when plotting a single series)
        yaxis : bool, optional
            Whether to show the yaxis and labels (default is False)
        ygrid : bool, optional
            Whether to show the yaxis grid (default is False)
        xgrid : bool, optional
            Whether to show the xaxis grid (default is True)
        range_tool : bool, optional
            Show the the range slider tool (default is True)
        height : int, optional
            The height of the plot figure
            (the default is auto-calculated height)
        width : int, optional
            The width of the plot figure (the default is 900)
        color : str
            Default series color (default is "navy")
        overlay_data : pd.DataFrame:
            A second dataframe to plot as a different series.
        overlay_color : str
            Overlay series color (default is "green")
        ref_events : pd.DataFrame, optional
            Add references line/label using the event times in the dataframe.
            (the default is None)
        ref_time_col : str, optional
            Add references line/label using the this column in `ref_events`
            for the time value (x-axis).
            (this defaults the value of the `time_column` parameter or 'TimeGenerated'
            `time_column` is None)
        ref_col : str, optional
            The column name to use for the label from `ref_events`
            (the default is None)
        ref_times : List[Tuple[datetime, str]], optional
            Add one or more reference line/label using (the default is None)

        Returns
        -------
        LayoutDOM
            The bokeh plot figure.

        """
        return self._display_timeline(data=self._df, **kwargs)

    # pylint: disable=invalid-name
    def plot_values(self, value_col: str = None, **kwargs) -> LayoutDOM:
        """
        Display a timeline of events.

        Parameters
        ----------
        time_column : str, optional
            Name of the timestamp column
            (the default is 'TimeGenerated')
        value_col : str
            The column name holding the value to plot vertically
        source_columns : list, optional
            List of default source columns to use in tooltips
            (the default is None)

        Other Parameters
        ----------------
        x : str, optional
            alias of `time_column`
        y : str, optional
            alias of `value_col`
        title : str, optional
            Title to display (the default is None)
        ref_event : Any, optional
            Add a reference line/label using the alert time (the default is None)
        ref_time : datetime, optional
            Add a reference line/label using `ref_time` (the default is None)
        ref_label : str, optional
            A label for the `ref_event` or `ref_time` reference item
        group_by : str
            (where `data` is a DataFrame)
            The column to group timelines on
        legend: str, optional
            "left", "right", "inline" or "none"
            (the default is to show a legend when plotting multiple series
            and not to show one when plotting a single series)
        yaxis : bool, optional
            Whether to show the yaxis and labels
        range_tool : bool, optional
            Show the the range slider tool (default is True)
        height : int, optional
            The height of the plot figure
            (the default is auto-calculated height)
        width : int, optional
            The width of the plot figure (the default is 900)
        color : str
            Default series color (default is "navy"). This is overridden by
            automatic color assignments if plotting a grouped chart
        kind : Union[str, List[str]]
            one or more glyph types to plot., optional
            Supported types are "circle", "line" and "vbar" (default is "vbar")
        ref_events : pd.DataFrame, optional
            Add references line/label using the event times in the dataframe.
            (the default is None)
        ref_time_col : str, optional
            Add references line/label using the this column in `ref_events`
            for the time value (x-axis).
            (this defaults the value of the `time_column` parameter or 'TimeGenerated'
            `time_column` is None)
        ref_col : str, optional
            The column name to use for the label from `ref_events`
            (the default is None)
        ref_times : List[Tuple[datetime, str]], optional
            Add one or more reference line/label using (the default is None)

        Returns
        -------
        LayoutDOM
            The bokeh plot figure.

        """
        return self._display_timeline_values(
            data=self._df, value_col=value_col, **kwargs
        )

    def plot_duration(
        self,
        group_by: Union[Iterable[str], str],
        time_column: str = "TimeGenerated",
        end_time_column: Optional[str] = None,
        **kwargs,
    ) -> LayoutDOM:  # noqa: C901, MC0001
        """
        Display a duration timeline of events grouped by one or more columns.

        Parameters
        ----------
        group_by : Union[Iterable[str], str]
            The column name or iterable of column names to group the data by.
        time_column : str
            Primary time column - will be used to calculate the
            start time of the duration for each group.
            If `end_time_column` is not specified it will also be used to
            calculate the end time.
        end_time_column : Optional[str]
            If supplied, it will be used to calculate the end time
            of the duration for each group.

        Other Parameters
        ----------------
        title : str, optional
            Title to display (the default is None)
        ylabel_cols : Optional[Iterable[str]], optional
            The subset of the group columns to use for the y-axis labels.
        yaxis : bool, optional
            Whether to show the yaxis and labels
        range_tool : bool, optional
            Show the the range slider tool (default is True)
        source_columns : list, optional
            List of default source columns to use in tooltips
            (the default is None)
        height : int, optional
            The height of the plot figure
            (the default is auto-calculated height)
        width : int, optional
            The width of the plot figure (the default is 900)
        color : str
            Default series color (default is "navy")
        ref_events : pd.DataFrame, optional
            Add references line/label using the event times in the dataframe.
            (the default is None)
        ref_col : str, optional
            The column name to use for the label from `ref_events`
            (the default is None)
        ref_times : List[Tuple[datetime, str]], optional
            Add one or more reference line/label using (the default is None)

        Returns
        -------
        LayoutDOM
            The bokeh plot figure.

        """
        return self._display_timeline_duration(
            data=self._df,
            group_by=group_by,
            time_column=time_column,
            end_time_column=end_time_column,
            **kwargs,
        )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Placeholder for old entity_schema module."""

from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"

# pylint: disable=wildcard-import, unused-wildcard-import
from ..datamodel.entities import *  # noqa: F403, F401



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""MSTIC Anomalous Sequence Modelling Utilities."""

from ...._version import VERSION

__version__ = VERSION



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Threatintelligence Entity class."""
from typing import Any, Dict, Mapping

from ..._version import VERSION
from ...common.utility import export
from .entity import Entity

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=invalid-name


@export
class UnknownEntity(Entity):
    """Generic Entity class."""

    def __init__(self, src_entity: Mapping[str, Any] = None, **kwargs):
        """
        Create a new instance of the entity type.

            :param src_entity: instantiate entity using properties of src entity
            :param kwargs: key-value pair representation of entity
        """
        super().__init__(src_entity=src_entity, **kwargs)
        if src_entity:
            self.__dict__.update(src_entity)

    @property
    def description_str(self) -> str:
        """Return Entity Description."""
        return "OtherEntity"

    @property
    def name_str(self) -> str:
        """Return Entity Name."""
        return self.__class__.__name__

    _entity_schema: Dict[str, Any] = {
        "TimeGenerated": None,
        "StartTime": None,
        "EndTime": None,
    }



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Entity enumerations."""
from enum import Enum
from typing import Dict, Type

from ..._version import VERSION
from ...common.utility import export

__version__ = VERSION
__author__ = "Ian Hellen"


ENTITY_ENUMS: Dict[str, Type] = {}


# pylint: disable=invalid-name
@export
class Algorithm(Enum):
    """FileHash Algorithm Enumeration."""

    Unknown = 0
    MD5 = 1
    SHA1 = 2
    SHA256 = 3
    SHA256AC = 4


@export
class ElevationToken(Enum):
    """ElevationToken enumeration."""

    Default = 0
    Full = 1
    Limited = 2


@export
class OSFamily(Enum):
    """OSFamily enumeration."""

    Linux = 0
    Windows = 1


@export
class RegistryHive(Enum):
    """RegistryHive enumeration."""

    # <summary>HKEY_LOCAL_MACHINE</summary>
    HKEY_LOCAL_MACHINE = 0
    # <summary>HKEY_CLASSES_ROOT</summary>
    HKEY_CLASSES_ROOT = 1
    # <summary>HKEY_CURRENT_CONFIG</summary>
    HKEY_CURRENT_CONFIG = 2
    # <summary>HKEY_USERS</summary>
    HKEY_USERS = 3
    # <summary>HKEY_CURRENT_USER_LOCAL_SETTINGS</summary>
    HKEY_CURRENT_USER_LOCAL_SETTINGS = 4
    # <summary>HKEY_PERFORMANCE_DATA</summary>
    HKEY_PERFORMANCE_DATA = 5
    # <summary>HKEY_PERFORMANCE_NLSTEXT</summary>
    HKEY_PERFORMANCE_NLSTEXT = 6
    # <summary>HKEY_PERFORMANCE_TEXT</summary>
    HKEY_PERFORMANCE_TEXT = 7
    # <summary>HKEY_A</summary>
    HKEY_A = 8
    # <summary>HKEY_CURRENT_USER</summary>
    HKEY_CURRENT_USER = 9

    @property
    def short_name(self) -> str:
        """Return the key shortname."""
        # pylint: disable=no-member
        return f"HK{''.join([n[0] for n in self.name.split('_')[1:]])}"
        # pylint: enable=no-member


ENTITY_ENUMS = {
    Algorithm.__name__: Algorithm,
    ElevationToken.__name__: ElevationToken,
    OSFamily.__name__: OSFamily,
    RegistryHive.__name__: RegistryHive,
}



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Url Entity class."""
from typing import Any, Dict, Mapping, Optional

from ..._version import VERSION
from ...common.utility import export
from ...sectools.domain_utils import url_components
from .entity import Entity

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=invalid-name


@export
class Url(Entity):
    """
    URL Entity.

    Attributes
    ----------
    Url : str
        The URL
    DetonationVerdict : str
        The verdict of the URL detection

    """

    ID_PROPERTIES = ["Url"]

    def __init__(
        self,
        src_entity: Mapping[str, Any] = None,
        src_event: Mapping[str, Any] = None,
        **kwargs,
    ):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing entity or
            other mapping object that implements entity properties.
            (the default is None)
        src_event : Mapping[str, Any], optional
            Create entity from event properties
            (the default is None)

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.

        """
        self.Url: Optional[str] = None
        self.DetonationVerdict: Optional[str] = None
        super().__init__(src_entity=src_entity, **kwargs)
        if src_event:
            self._create_from_event(src_event)

    def _create_from_event(self, src_event):
        self.Url = src_event["Url"]
        if "AdditionalData" in src_event:
            self.DetonationVerdict = src_event["AdditionalData"]["DetonationVerdict"]

    @property
    def description_str(self) -> str:
        """Return Entity Description."""
        return self.Url or super().description_str

    @property
    def name_str(self) -> str:
        """Return Entity Name."""
        return self.Url or self.__class__.__name__

    def __getitem__(self, key: str):
        """Allow property get using dictionary key syntax."""
        if self.Url:
            val = url_components(self.Url).get(key)
            if val:
                return val
        return super().__getitem__(key)

    def __getattr__(self, name: str):
        """Return the value of the named property 'name'."""
        if self.Url:
            val = url_components(self.Url).get(name)
            if val:
                return val
        return super().__getattr__(name)

    _entity_schema: Dict[str, Any] = {
        # Url (type System.String)
        "Url": None,
        "DetonationVerdict": None,
        "TimeGenerated": None,
        "StartTime": None,
        "EndTime": None,
    }



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Module for SQL to KQL Conversion.

This is an experiment conversion utility built to support a limited subset
of ANSI SQL.
It relies on moz_sql_parser (https://github.com/mozilla/moz-sql-parser)
to parse the SQL syntax tree. Some hacky additions have been done to
allow table renaming and support for a few SparkSQL operators such as
RLIKE.

For a more complete translation help with SQL to KQL see
https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/sqlcheatsheet

Known limitations
-----------------

- Does not support aggregate functions in SELECT with no GROUP BY clause
- Does not support IN, EXISTS, HAVING operators
- Only partial support for AS naming (should work in SELECT expressions)

"""
import random
import re
from typing import List, Tuple, Any, Union, Dict, Optional

from ..common.exceptions import MsticpyImportExtraError

try:
    import moz_sql_parser
    from moz_sql_parser import parse
except ImportError as imp_err:
    raise MsticpyImportExtraError(
        "Cannot use this feature without moz_sql_parser installed",
        title="Error importing moz_sql_parser for sql_to_kql",
        extra="sql",
    ) from imp_err


from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"

_DEBUG = False

SPARK_KQL_FUNC_MAP = {
    "avg": ("avg", None, None),
    "base64": ("base64_encode_tostring", None, None),
    "concat": ("strcat", None, None),
    "count": ("count", None, None),
    "if": ("iif", None, None),
    "iif": ("iif", None, None),
    "ifnull": (None, None, "iif(isnull({p0}), {p1}, {p0}))"),
    "in": ("in", None, None),
    "instr": ("indexof", None, None),
    "int": ("toint", None, None),
    "isnotnull": ("isnotnull", None, None),
    "isnull": ("isnull", None, None),
    "left": ("NA", None, None),
    "length": ("strlen", None, None),
    "like": ("NA", None, None),
    "locate": ("indexof", None, None),
    "lower": ("tolower", None, None),
    "lcase": ("tolower", None, None),
    "ltrim": ("trim_start", None, None),
    "max": ("max", None, None),
    "mean": ("mean", None, None),
    "min": ("min", None, None),
    "position": ("indexof", None, None),
    "regexp_extract": ("extract", "{p1}, {p0}", None),  # swap params 0, 1
    "replace": ("replace", None, None),
    "reverse": ("reverse", None, None),
    "rtrim": ("trim_end", None, None),
    "split": ("split", None, None),
    "string": ("tostring", None, None),
    "str": ("tostring", None, None),
    "substr": ("substring", None, None),
    "substring": ("substring", None, None),
    "sum": ("sum", None, None),
    "todatetime": ("to_date", None, None),
    "translate": ("translate", None, None),
    "trim": ("trim", None, None),
    "unbase64": ("base64_decode_tostring", None, None),
    "upper": ("toupper", None, None),
}


AND = "and"
AS = "as"
ASC = "asc"
BETWEEN = "between"
CASE = "case"
COLLATE_NOCASE = "collate nocase"
CROSS_JOIN = "cross join"
DESC = "desc"
DISTINCT = "distinct"
ELSE = "else"
END = "end"
FROM = "from"
FULL_JOIN = "full join"
FULL_OUTER_JOIN = "full outer join"
GROUP_BY = "groupby"
HAVING = "having"
IN = "in"
INNER_JOIN = "inner join"
IS = "is"
IS_NOT = "is not"
JOIN = "join"
LEFT_JOIN = "left join"
LEFT_OUTER_JOIN = "left outer join"
LIKE = "like"
LIMIT = "limit"
NOT = "not"
NOT_BETWEEN = "not between"
NOT_IN = "not in"
NOT_LIKE = "not like"
OFFSET = "offset"
ON = "on"
OR = "or"
ORDER_BY = "orderby"
RIGHT_JOIN = "right join"
RIGHT_OUTER_JOIN = "right outer join"
SELECT = "select"
THEN = "then"
UNION = "union"
UNION_ALL = "union all"
USING = "using"
WHEN = "when"
WHERE = "where"
WITH = "with"

# override/add keywords
RLIKE = "rlike"


JOIN_KEYWORDS = {
    FULL_JOIN: "outer",
    FULL_OUTER_JOIN: "outer",
    INNER_JOIN: "inner",
    JOIN: "inner",
    LEFT_JOIN: "left",
    LEFT_OUTER_JOIN: "left",
    RIGHT_JOIN: "right",
    RIGHT_OUTER_JOIN: "right",
    CROSS_JOIN: "CROSS_JOIN_TODO",
}
JOIN_KEYWORDS = {kw.replace("_", " "): kql for kw, kql in JOIN_KEYWORDS.items()}


BINARY_OPS = {val: key for key, val in moz_sql_parser.keywords.binary_ops.items()}
BINARY_OPS["eq"] = "=="
BINARY_OPS["neq"] = "!="
BINARY_OPS["nin"] = "!in"
BINARY_OPS["rlike"] = "not matches regex"
BINARY_OPS["nlike"] = "not matches regex"
BINARY_OPS["concat"] = "+"
BINARY_OPS["is"] = "=="
BINARY_OPS["is not"] = "!="


REMAPPED_KEYWORDS = {"RLIKE": "LIKE"}


# noqa: MC0001


def sql_to_kql(sql: str, target_tables: Dict[str, str] = None) -> str:
    """Parse SQL and return KQL equivalent."""
    # ensure literals are surrounded by single quotes
    sql = _single_quote_strings(sql)

    # replace table names
    if target_tables:
        for table in target_tables:
            sql = sql.replace(table, target_tables[table])
    # replace keywords
    sql = _remap_kewords(sql)
    parsed_sql = parse(sql)
    query_lines = _parse_query(parsed_sql)
    return "\n".join(line for line in query_lines if line.strip())


def _parse_query(parsed_sql: Dict[str, Any]) -> List[str]:  # noqa: MC0001
    """Translate query or subquery."""
    query_lines: List[str] = []
    if isinstance(parsed_sql, str):
        return [parsed_sql]
    if FROM in parsed_sql:
        _process_from(parsed_sql[FROM], query_lines)
    if WHERE in parsed_sql:
        query_lines.append(f"| where {_parse_expression(parsed_sql[WHERE])}")

    if GROUP_BY in parsed_sql:
        _process_group_by(parsed_sql, query_lines)
        # Get rid of the SELECT statement since we've processed it in
        # the groupby
        parsed_sql.pop(SELECT)

    distinct_select: List[Dict[str, Any]] = []
    if SELECT in parsed_sql:
        distinct_select, expr_list = _is_distinct(parsed_sql[SELECT])
        _process_select(parsed_sql[SELECT], expr_list, query_lines)
    if ORDER_BY in parsed_sql:
        query_lines.append(f"| order by {_create_order_by(parsed_sql[ORDER_BY])}")
    if distinct_select:
        query_lines.append(
            f"| distinct {', '.join(_create_distinct_list(distinct_select))}"
        )
    if LIMIT in parsed_sql:
        query_lines.append(f"| limit {parsed_sql[LIMIT]}")
    if UNION in parsed_sql:
        union_subquery = {UNION_ALL: parsed_sql[UNION]}
        query_lines.extend(_parse_query(union_subquery))
        query_lines.append("| distinct *")
    if UNION_ALL in parsed_sql:
        union_l_expr = "\n".join(_parse_query(parsed_sql[UNION_ALL][0]))
        query_lines.append(union_l_expr)
        union_r_expr = "\n  ".join(_parse_query(parsed_sql[UNION_ALL][1]))
        query_lines.append(f"| union ({union_r_expr}\n)")

    return query_lines


def _process_from(
    from_expr: Union[List[Dict[str, Any]], Dict[str, Any], str], query_lines: List[str]
):
    """Process FROM clause."""
    if isinstance(from_expr, dict) and UNION in from_expr:
        query_lines.extend(_parse_query(from_expr))
    elif isinstance(from_expr, dict):
        query_lines.extend(_parse_query(from_expr["value"]))
    elif isinstance(from_expr, str):
        query_lines.append((from_expr))
        return
    elif isinstance(from_expr, list):
        for from_item in from_expr:
            if isinstance(from_item, str):
                query_lines.append((from_item))
            elif isinstance(from_item, dict) and "value" in from_item:
                query_lines.extend(_parse_query(from_item.get("value")))  # type: ignore

    join_expr = from_expr if isinstance(from_expr, list) else [from_expr]
    join_list = _get_join_list(join_expr)
    for join_item in join_list:
        join_line = _parse_join(join_item)
        if join_line:
            query_lines.append(join_line)


def _process_select(
    parsed_sql: Dict[str, Any],
    expr_list: Union[List[Dict[Any, Any]], Dict[Any, Any]],
    query_lines: List[str],
):
    """Process SELECT clause."""
    # Expressions
    if parsed_sql == "*":
        return
    _db_print(expr_list, type(expr_list))
    select_list = expr_list if isinstance(expr_list, list) else [expr_list]
    select_list = _get_expr_list(select_list)
    project_items = []
    extend_items = []
    for item in select_list:
        value = _parse_expression(item["value"])
        name = item.get("name")
        if value != item["value"]:
            # if this isn't a simple rename - add to extend
            name = name or _gen_expr_name(item["value"])
            extend_items.append(f"{name} = {value}")
            project_items.append(name)
        else:
            if name:
                project_items.append(f"{name} = {value}")
            else:
                project_items.append(value)
    if extend_items:
        query_lines.append(f"| extend {', '.join(extend_items)}")
    if project_items:
        query_lines.append(f"| project {', '.join(project_items)}")


def _gen_expr_name(value):
    """Generate random expression name."""
    pref = "expr"
    if isinstance(value, str):
        return value
    if isinstance(value, dict):
        first_val = next(iter(value.values()))
        if isinstance(first_val, str):
            return first_val
        # prob a function so base the name on that
        pref = next(iter(value.keys()))
    # otherwise just generate a rand value as suffix
    suffix = str(random.randint(1000, 9999))  # nosec
    return f"{pref}_{suffix}"


def _get_expr_list(expr_list):
    if (
        isinstance(expr_list, list)
        and len(expr_list) == 1
        and isinstance(expr_list[0], dict)
        and isinstance(expr_list[0].get("value"), list)
    ):
        return expr_list[0]["value"]
    return expr_list


def _get_expr_value(expr_val):
    if isinstance(expr_val, dict) and "value" in expr_val:
        return expr_val["value"]
    if isinstance(expr_val, list):
        return _get_expr_list(expr_val)
    return expr_val


def _process_group_by(parsed_sql: Dict[str, Any], query_lines: List[str]):
    """Process GROUP BY clause."""
    group_by_expr = parsed_sql[GROUP_BY]
    group_by_expr = (
        group_by_expr if isinstance(group_by_expr, list) else [group_by_expr]
    )
    by_clause = ", ".join(val["value"] for val in group_by_expr if val.get("value"))

    _, expr_list = _is_distinct(parsed_sql[SELECT])
    group_by_expr_list = []
    expr_list = _get_expr_value(expr_list)
    for expr in expr_list:
        name_expr = ""
        if "name" in expr:
            name_expr = f"{expr.get('name')} = "
        if isinstance(expr.get("value"), str):
            group_by_expr_list.append(f"{name_expr}any({expr['value']})")
        else:
            group_by_expr_list.append(
                f"{name_expr}{_parse_expression(expr.get('value'))}"
            )
    query_lines.append(f"| summarize {', '.join(group_by_expr_list)} by {by_clause}")


# pylint: disable=too-many-return-statements, too-many-branches
def _parse_expression(expression):  # noqa: MC0001
    """Return parsed expression."""
    if _is_literal(expression)[0]:
        return _quote_literal(expression)
    if not isinstance(expression, dict):
        return expression
    if AND in expression:
        return "\n  and ".join([_parse_expression(expr) for expr in expression[AND]])
    if OR in expression:
        return "\n  or ".join([_parse_expression(expr) for expr in expression[OR]])
    if NOT in expression:
        return f" not ({_parse_expression(expression[NOT])})"
    if BETWEEN in expression:
        args = expression[BETWEEN]
        betw_expr = f"{_parse_expression(args[1])} .. {_parse_expression(args[2])}"
        return f"{args[0]} between ({betw_expr})"
    if NOT_BETWEEN in expression:
        args = expression[NOT_BETWEEN]
        betw_expr = f"{_parse_expression(args[1])} .. {_parse_expression(args[2])}"
        return f"{args[0]} not between ({betw_expr})"
    if IN in expression or NOT_IN in expression:
        sql_op = IN if IN in expression else NOT_IN
        kql_op = IN if IN in expression else "!in"
        args = expression[sql_op]

        right = _quote_literal(args[1])
        if isinstance(right, list):
            _db_print(args[1])
            arg_list = ", ".join([str(_parse_expression(l_item)) for l_item in right])
            return f"{args[0]} {kql_op} ({arg_list})"
        sub_query = "\n".join(_parse_query(right))
        return f"{args[0]} {kql_op} ({sub_query})"

    # Handle other operators
    oper = list(expression.keys())[0] if expression else None
    if oper in BINARY_OPS:
        right = _parse_expression(expression[oper][1])
        left = _parse_expression(expression[oper][0])
        return f"{left} {BINARY_OPS[oper]} {right}"
    if LIKE in expression:
        return _process_like(expression)

    # For everything else, assume it's a function
    if expression:
        func, operand = next(iter(expression.items()))
        #         set_trace()
        return _map_func(func, operand)
    return "EXPRESSION {expression} not resolved."


# pylint: enable=too-many-return-statements, too-many-branches


def _map_func(func: str, *args) -> str:
    """Return KQL function for SQL function."""
    func = func.lower().strip()
    args_dict = {f"p{idx}": arg for idx, arg in enumerate(args)}
    def_arg_fmt = ", ".join(f"{{{arg}}}" for arg in args_dict)
    if func not in SPARK_KQL_FUNC_MAP:
        func_fmt = f"{func}({def_arg_fmt}) // WARNING unmapped function\n"
        return func_fmt.format(**args_dict)
    func_map = SPARK_KQL_FUNC_MAP[func]

    if (
        func == "count"
        and isinstance(args[0], dict)
        and next(iter(args[0])) == "distinct"
    ):
        func_arg = _get_expr_value(args[0]["distinct"])
        return f"dcount({func_arg})"

    if not func_map[1] and not func_map[2]:
        func_fmt = f"{func_map[0]}({def_arg_fmt})"
        return func_fmt.format(**args_dict)
    if func_map[1]:
        func_fmt = f"{func_map[0]}({func_map[1]})"
        return func_fmt.format(**args_dict)
    if func_map[2]:
        func_fmt = f"{func_map[2]}"
        return func_fmt.format(**args_dict)
    raise ValueError(f"Could not map function or args {func}{args_dict}")


def _quote_literal(expr: Union[str, List[str], Any]) -> Any:
    """Quote string if it is a literal."""
    literal, expr = _is_literal(expr)
    if not literal:
        return expr
    if isinstance(expr, str):
        return _quote(expr)
    if isinstance(expr, list):
        return [_quote(memb) for memb in expr]
    return expr


def _is_literal(expr: Union[Dict[str, Any], Any]) -> Tuple[bool, Any]:
    """Check if literal string."""
    if isinstance(expr, dict) and "literal" in expr:
        return True, expr["literal"]
    return False, expr


def _quote(expr: str) -> str:
    """Quote a string, if not already quoted."""
    if expr.startswith("'") and expr.endswith("'"):
        return expr
    return f"'{expr}'"


def _single_quote_strings(sql: str) -> str:
    """Replace unquoted double-quotes with single-quotes."""
    return re.sub(r"(?<![\\])\"", "'", sql)


def _remap_kewords(sql: str) -> str:
    """Replace keywords in source SQL statement."""
    for repl_kw, mapped_kw in REMAPPED_KEYWORDS.items():
        sql = re.sub(f"\\s{repl_kw}\\s", f" {mapped_kw} ", sql)
    return sql


def _is_distinct(
    select_list: Union[Dict[str, Any], List[Dict[str, Any]]]
) -> Tuple[List[Dict[str, Any]], List[Dict[str, Any]]]:
    """Check for DISTINCT in SELECT clause."""
    select_list_out = []
    dist_list: List[Dict[str, Any]] = []
    dist_dict = _get_expr_value(select_list)

    if isinstance(dist_dict, dict) and DISTINCT in dist_dict:
        dist_list = dist_dict.pop(DISTINCT)
        # Keep distinct items in the select list
        return dist_list, dist_list

    if isinstance(select_list, dict):
        select_list = [select_list]
    if isinstance(select_list, list):
        for expr in select_list:
            _db_print(expr)
            if "value" in expr and DISTINCT in expr["value"]:
                dist_list.append({"value": _get_expr_value(expr["value"][DISTINCT])})
            select_list_out.append(expr)
    return dist_list, select_list_out


def _format_order_item(item: Dict[str, Any]) -> str:
    """Return ORDER BY item with sort direction."""
    if "sort" in item:
        return f"{item['value']} {item['sort'].lower()}"
    return f"{item['value']}"


def _get_join_list(parsed_sql: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """Return list of JOIN sub-expressions."""
    if not isinstance(parsed_sql, list):
        return []
    join_list = []
    for from_source in parsed_sql:
        if not isinstance(from_source, dict):
            continue
        join = JOIN_KEYWORDS & from_source.keys()
        if join:
            join_list.append(from_source)
    return join_list


def _rewrite_table_refs(join_expr: Union[Any, str, List], table_expr: str) -> str:
    """Rewrite dotted prefixes."""
    p_expr = _parse_expression(join_expr)
    prefixes = set(re.findall(r"(\w+)\.", p_expr))
    if not prefixes:
        return p_expr
    if f"{table_expr}" in prefixes:
        p_expr = p_expr.replace(f"{table_expr}.", "$right.")
        prefixes.remove(f"{table_expr}")
    for prefix in prefixes:
        p_expr = p_expr.replace(
            f"{prefix}.",
            "$right." if prefix.casefold() == table_expr.casefold() else "$left.",
        )
    return p_expr


def _parse_join(join_expr) -> Optional[str]:
    """Return translated JOIN expression."""
    join_type_set = JOIN_KEYWORDS & join_expr.keys()
    if not join_type_set:
        return None
    join_type = join_type_set.pop()
    table_expr = join_expr[join_type]
    kql_join_type = JOIN_KEYWORDS[join_type]
    if "value" in table_expr and "select" in table_expr["value"]:
        table_expr = table_expr["value"]

    p_table_expr = "\n  ".join(_parse_query(table_expr))
    if "name" in join_expr[join_type]:
        table_name = join_expr[join_type]["name"]
    else:
        table_name = p_table_expr.split(" ", maxsplit=1)[0].strip()
    on_expr = _parse_expression(join_expr["on"])
    on_expr = _rewrite_table_refs(on_expr, table_name)
    _db_print(table_expr, kql_join_type, p_table_expr)

    return f"| join kind={kql_join_type} ({p_table_expr}) on {on_expr}"


def _process_like(expression: Dict[str, Any]) -> str:
    """Process Like clause."""
    left = _parse_expression((expression[LIKE][0]))
    literal, right = _is_literal(expression[LIKE][1])
    if not (literal and isinstance(right, str)):
        raise ValueError(
            f"Right side operand {right} isn't usable in LIKE expression",
            f"{left} LIKE {right}",
        )
    if re.match("^[^%_]+[%_]$", right):
        oper = "startswith"
        right = right.replace("%", "").replace("_", "")
    elif re.match("^[%_][^%_]+$", right):
        oper = "endswith"
        right = right.replace("%", "").replace("_", "")
    elif re.match("^[%_][^%_]+[%_]$", right):
        oper = "contains"
        right = right.replace("%", "").replace("_", "")
    else:
        oper = "matches regex"
        right = right.replace("_", ".").replace("%", ".*")
    right = _quote(right)
    return f"{left} {oper} {right}"


def _create_distinct_list(distinct_select):
    distinct_list = []
    for distinct_item in distinct_select:
        if "name" in distinct_item:
            distinct_list.append(distinct_item["name"])
        else:
            val = _parse_expression(_get_expr_value(distinct_item))
            if val != distinct_item.get("value"):
                # If value was a complex expression we can't use
                # it directly so just revert to distinct *
                distinct_list = ["*"]
                break
            distinct_list.append(val)
    return distinct_list


def _create_order_by(order_by):
    if isinstance(order_by, list):
        return ", ".join(_format_order_item(item) for item in order_by)
    return _format_order_item(order_by)


def _db_print(*args, **kwargs):
    if _DEBUG:
        print(*args, **kwargs)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Data query definition reader."""
from typing import Tuple, Dict, Iterable, Any
from pathlib import Path
import yaml

from .query_defns import DataEnvironment
from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


def find_yaml_files(source_path: str, recursive: bool = False) -> Iterable[Path]:
    """
    Return iterable of yaml files found in `source_path`.

    Parameters
    ----------
    source_path : str
        The source path to search in.
    recursive : bool, optional
        Whether to recurse through subfolders.
        By default False

    Returns
    -------
    Iterable[str]
        File paths of yanl files found.

    """
    recurse_pfx = "**/" if recursive else ""
    file_glob = Path(source_path).glob(f"{recurse_pfx}*.yaml")
    for file_path in file_glob:
        if not file_path.is_file():
            continue
        yield file_path


def read_query_def_file(query_file: str) -> Tuple[Dict, Dict, Dict]:
    """
    Read a yaml data query definition file.

    Parameters
    ----------
    query_file : str
        Path to yaml query defintion file

    Returns
    -------
    Tuple[Dict, Dict, Dict]
        Tuple of dictionaries.
        sources - dictionary of query definitions
        defaults - the default parameters from the file
        metadata - the global metadata from the file

    """
    data_map = None
    with open(query_file, "r", encoding="utf-8") as f_handle:
        # use safe_load instead load
        data_map = yaml.safe_load(f_handle)

    validate_query_defs(query_def_dict=data_map)

    defaults = data_map.get("defaults", {})
    sources = data_map.get("sources", {})
    metadata = data_map.get("metadata", {})

    return sources, defaults, metadata


def validate_query_defs(query_def_dict: Dict[str, Any]) -> bool:
    """
    Validate content of query definition.

    Parameters
    ----------
    query_def_dict : dict
        Dictionary of query definition yaml file contents.

    Returns
    -------
    bool
        True if validation succeeds.

    Raises
    ------
    ValueError
        The validation failure reason is returned in the
        exception message (arg[0])

    """
    # verify that sources and metadata are in the data dict
    if "sources" not in query_def_dict or not query_def_dict["sources"]:
        raise ValueError("Imported file has no sources defined")
    if "metadata" not in query_def_dict or not query_def_dict["metadata"]:
        raise ValueError("Imported file has no metadata defined")

    # data_environments and data_families must be defined at with at least
    # one value
    _validate_data_categories(query_def_dict)

    return True


def _validate_data_categories(query_def_dict: Dict):
    if (
        "data_environments" not in query_def_dict["metadata"]
        or not query_def_dict["metadata"]["data_environments"]
    ):
        raise ValueError("Imported file has no data_environments defined")

    for env in query_def_dict["metadata"]["data_environments"]:
        if not DataEnvironment.parse(env):
            raise ValueError(
                f"Unknown data environment {env} in metadata. ",
                "Valid values are\n",
                ", ".join(e.name for e in DataEnvironment),
            )

    if (
        "data_families" not in query_def_dict["metadata"]
        or not query_def_dict["metadata"]["data_families"]
    ):
        raise ValueError("Imported file has no data families defined")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Bokeh matrix plot."""
import math
from typing import List, Optional, Union

import attr
from bokeh.io import output_notebook, show, reset_output
from bokeh.plotting import figure
from bokeh.models import (
    HoverTool,
    ColumnDataSource,
    LayoutDOM,
)
import numpy as np
import pandas as pd

from ..common.utility import check_kwargs
from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


@attr.s(auto_attribs=True)
class PlotParams:
    """Plot params for time_duration."""

    title: Optional[str] = "Interaction Plot"
    x: Optional[str] = None
    x_col: Optional[str] = None
    y: Optional[str] = None
    y_col: Optional[str] = None
    intersect: bool = False
    height: int = 700
    width: int = 900
    color: str = "red"
    value_col: Optional[str] = None
    dist_count: bool = False
    log_size: bool = False
    invert: bool = False
    sort: Optional[Union[str, bool]] = None
    sort_x: Optional[Union[str, bool]] = None
    sort_y: Optional[Union[str, bool]] = None
    hide: bool = False
    font_size: Optional[int] = None
    max_label_font_size: int = 11

    @property
    def x_column(self) -> Optional[str]:
        """Return the current x column value."""
        return self.x or self.x_col

    @property
    def y_column(self) -> Optional[str]:
        """Return the current y column value."""
        return self.y or self.y_col

    @classmethod
    def field_list(cls) -> List[str]:
        """Return field names as a list."""
        return list(attr.fields_dict(cls).keys())


def plot_matrix(data: pd.DataFrame, **kwargs) -> LayoutDOM:
    """
    Plot data as an intersection matrix.

    Parameters
    ----------
    data : pd.DataFrame
        The data to plot.
    x : str
        Column to plot on the x (horizontal) axis
    x_col : str
        Alias for 'x'
    y : str
        Column to plot on the y (vertical) axis
    y_col : str
        Alias for 'y'
    title : str, optional
        Custom title, default is 'Intersection plot'
    value_col : str, optional
        Column from the DataFrame used to size the intersection points.
    dist_count : bool, optional
        Calculates a count of distinct values (from `value_col`) and uses
        this to size the intersection points.
        Requires `value_col` to be specified.
    log_size : bool, optional
        Takes the log of the size value before calculating the intersection
        display point size.
        Can be combined with `invert`.
    invert : bool, optional
        Takes the inverse of the size value as the basis for calculating
        the intersection display point size. This is useful for highlighting
        rare interactions.
        Can be combined with `log_size`.
    intersect : bool, optional
        Plots points of a fixed size, rather than using a sizing value. This
        is useful for just showing the presence/absence of an interaction.
    height : int, optional
        The plot height. Default is 700
    width : int
        The plot width. Default is 900
    color : str
        The color of the plotted points, default is "red"
    sort : Union[str, bool], optional
        Sorts the labels of both axes, default is None.
        Acceptable values are:
        'asc' (or string starting with 'asc') - Sort ascending
        'desc' (or string starting with 'asc') - Sort descending
        False or None (no sort)
        True  - Sort ascending
    sort_x : str, optional
        Sorts the labels of the x axis (takes precedence over `sort`),
        default is None.
        Acceptable values are:
        'asc' (or string starting with 'asc') - Sort ascending
        'desc' (or string starting with 'asc') - Sort descending
        False or None (no sort)
        True  - Sort ascending
    sort_y : str, optional
        Sorts the labels of the y axis (takes precedence over `sort`),
        default is None.
        Acceptable values are:
        'asc' (or string starting with 'asc') - Sort ascending
        'desc' (or string starting with 'asc') - Sort descending
        False or None (no sort)
        True  - Sort ascending
    hide : bool, optional
        Creates and returns but does not display the plot, default
        is False.
    font_size : int, optional
        Manually specify the font size for axis labels, in points,
        the default is to automatically calculate a size based on the
        number of items in each axis.
    max_label_font_size : int, optional
        The maximum size, in points, of the X and Y labels, default is 11.


    Returns
    -------
    LayoutDOM
        The Bokeh plot

    """
    # Process/extract parameters
    check_kwargs(kwargs, PlotParams.field_list())
    param = PlotParams(**kwargs)

    if not param.x_column or not param.y_column:
        raise ValueError("Must supply `x` and `y` column parameters.")

    reset_output()
    output_notebook()

    plot_data = _prep_data(data, param)

    x_range = _sort_labels(plot_data, param.x_column, param.sort_x or param.sort)
    y_range = _sort_labels(
        plot_data, param.y_column, param.sort_y or param.sort, invert=True
    )

    # Rescale the size so that it matches the graph
    max_size = plot_data["size"].max()
    plot_data["plt_size"] = plot_data["size"] * 10 / max_size
    source = ColumnDataSource(data=plot_data)

    plot = figure(
        title=param.title,
        plot_width=param.width,
        plot_height=param.height,
        x_range=x_range,
        y_range=y_range,
        tools=["wheel_zoom", "box_zoom", "pan", "reset", "save"],
        toolbar_location="above",
    )

    tool_tips = [
        (param.x_column, f"@{param.x_column}"),
        (param.y_column, f"@{param.y_column}"),
        ("value", "@size"),
    ]
    plot.add_tools(HoverTool(tooltips=tool_tips))

    if param.intersect:
        plot.circle_cross(
            x=param.x_column,
            y=param.y_column,
            source=source,
            fill_alpha=0.6,
            line_color=param.color,
            size=5,
        )
    else:
        plot.circle(
            x=param.x_column,
            y=param.y_column,
            source=source,
            fill_alpha=0.6,
            fill_color=param.color,
            size="plt_size",
        )
    _set_plot_params(plot)

    # Calculate appropriate font size for labels
    x_label_pt_size = param.font_size or max(
        5,
        min(
            param.max_label_font_size,
            int(param.width * 0.6 / plot_data[param.x_column].nunique()),
        ),
    )
    y_label_pt_size = param.font_size or max(
        5,
        min(
            param.max_label_font_size,
            int(param.height * 0.6 / plot_data[param.y_column].nunique()),
        ),
    )
    plot.xaxis.major_label_text_font_size = f"{x_label_pt_size}pt"
    plot.yaxis.major_label_text_font_size = f"{y_label_pt_size}pt"
    plot.xaxis.axis_label = param.x_column
    plot.yaxis.axis_label = param.y_column

    if not param.hide:
        show(plot)
    return plot


def _set_plot_params(plot):
    plot.title.text_font_size = "15pt"
    plot.outline_line_color = None
    plot.grid.grid_line_color = "navy"
    plot.grid.grid_line_alpha = 0.1
    plot.axis.axis_line_color = None
    plot.axis.major_tick_line_color = None
    plot.xaxis.major_label_orientation = math.pi / 2
    plot.xgrid.visible = True
    plot.ygrid.visible = True
    plot.axis.major_label_standoff = 0


def _sort_labels(data, column, sort_arg, invert=False):
    """Sort the labels if required."""
    if sort_arg:
        if isinstance(sort_arg, str):
            sort_order = sort_arg.casefold().startswith("asc")
        else:
            sort_order = True
        sort_order = not sort_order if invert else sort_order
        return data[column].sort_values(ascending=sort_order).unique().tolist()
    return data[column].unique().tolist()


def _prep_data(data: pd.DataFrame, param: PlotParams):
    """Process the data to create size column."""

    def _size_scale(value_series, log_size, invert):
        # local function to scale values
        if invert:
            # If invert, calculate inverse values on same
            # scale as input
            max_size = value_series.max()
            # min_size = value_series.min()
            value_series = (max_size) / value_series
        if log_size:
            # calc log of values, if requested
            return np.log(value_series)
        return value_series

    if param.value_col is None:
        # calculate a count of rows in each group
        other_cols = list(set(data.columns) - set([param.x_column, param.y_column]))
        if other_cols:
            count_col = other_cols[0]
        else:
            count_col = data.index.name or "index"
            data = data.reset_index()
        count_rows_df = (
            data[[param.x_column, param.y_column, count_col]]
            .groupby([param.x_column, param.y_column])
            .count()
            .rename(columns={count_col: "row_count"})
            .reset_index()
        )
        return count_rows_df.assign(
            size=_size_scale(count_rows_df.row_count, param.log_size, param.invert)
        )

    # if value column was specified, use that
    if param.dist_count:
        # If distinct count of values required, get nunique
        tmp_df = (
            data[[param.x_column, param.y_column, param.value_col]]
            .groupby([param.x_column, param.y_column])
            .nunique()
            .reset_index()
        )
    else:
        tmp_df = (
            data[[param.x_column, param.y_column, param.value_col]]
            .groupby([param.x_column, param.y_column])
            .sum()
            .reset_index()
        )
    return tmp_df.assign(
        size=lambda x: _size_scale(
            tmp_df[param.value_col], param.log_size, param.invert
        )
    )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Query helper definitions."""
from abc import ABC, abstractmethod
from enum import Enum
from typing import Union, List

import pandas as pd

from ..common.utility import export
from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=invalid-name
@export
class DataFamily(Enum):
    """
    Enumeration of data families.

    Used to identify which queries are relevant for which
    data sources.
    """

    Unknown = 0
    WindowsSecurity = 1
    LinuxSecurity = 2
    SecurityAlert = 3
    SecurityGraphAlert = 4
    LinuxSyslog = 5
    AzureNetwork = 6
    MDATP = 7
    Splunk = 8
    ResourceGraph = 9
    Sumologic = 10

    @classmethod
    def parse(cls, value: Union[str, int]) -> "DataFamily":
        """
        Convert string or int to enum.

        Parameters
        ----------
        value : Union[str, int]
            value to parse

        """
        if isinstance(value, cls):
            return value

        parsed_enum = cls.Unknown
        if isinstance(value, str):
            try:
                parsed_enum = cls[value]
            except KeyError:
                # match to value if case is incorrect
                # pylint: disable=no-member
                for e_name, e_val in cls.__members__.items():
                    if e_name.upper() == value.upper():
                        return e_val
                return cls.Unknown
                # pylint: enable=no-member
        if isinstance(value, int):
            try:
                parsed_enum = cls(value)
            except ValueError:
                parsed_enum = cls.Unknown
        return parsed_enum


@export
class DataEnvironment(Enum):
    """
    Enumeration of data environments.

    Used to identify which queries are relevant for which
    data sources.
    """

    Unknown = 0
    AzureSentinel = 1  # alias of LogAnalytics
    LogAnalytics = 1
    MSSentinel = 1
    Kusto = 2
    AzureSecurityCenter = 3
    SecurityGraph = 4
    MDE = 5  # alias of MDATP
    MDATP = 5
    LocalData = 6
    Splunk = 7
    Mordor = 8
    ResourceGraph = 9
    Sumologic = 10
    M365D = 11

    @classmethod
    def parse(cls, value: Union[str, int]) -> "DataEnvironment":
        """
        Convert string or int to enum.

        Parameters
        ----------
        value : Union[str, int]
            value to parse

        """
        if isinstance(value, cls):
            return value

        parsed_enum = cls.Unknown
        if isinstance(value, str):
            try:
                parsed_enum = cls[value]
            except KeyError:
                pass
        if isinstance(value, int):
            parsed_enum = cls(value)
        return parsed_enum


# pylint: disable=too-few-public-methods
@export
class QueryParamProvider(ABC):
    """
    Abstract type for QueryParamProvider.

    Method query_params must be overridden by derived classes.

    """

    @property
    @abstractmethod
    def query_params(self):
        """
        Return dict of query parameters.

        These parameters are sourced in the object
        implementing this method.

        Returns
        -------
            dict -- dictionary of query parameter values.

        """
        return {}


def ensure_df_datetimes(
    data: pd.DataFrame,
    columns: Union[str, List[str], None] = None,
    add_utc_tz: bool = True,
) -> pd.DataFrame:
    """
    Return dataframe with converted TZ-aware timestamps.

    Parameters
    ----------
    data : pd.DataFrame
        Input dataframe
    columns : Union[str, List[str], None], optional
        column (str) or list of columns to convert, by default None.
        If this parameter is not supplied then any column containing
        the substring "time" is used as a candidate for conversion.
    add_utc_tz: bool, optional
        If True any datetime columns in the `columns` parameter (
        (or default `'.*time.*'` columns) that are timezone-naive,
        will be converted to Timezone-aware timestamps marked as UTC.

    Returns
    -------
    pd.DataFrame
        Converted DataFrame.

    """
    if not columns:
        columns = list(data.filter(regex=".*[Tt]ime.*").columns)
    if isinstance(columns, str):
        columns = [columns]
    col_map = {
        col: "datetime64[ns, UTC]"
        for col in set(columns)
        if col in data.columns and not pd.api.types.is_datetime64_any_dtype(data[col])
    }
    converted_data = data.astype(col_map, errors="ignore")

    # Look for any TZ-naive columns in the list
    if add_utc_tz:
        localize_cols = {
            col for col in columns if col in data.select_dtypes("datetime")
        }
        for col in localize_cols:
            converted_data[col] = converted_data[col].dt.tz_localize(
                "UTC", ambiguous="infer", nonexistent="shift_forward"
            )
    return converted_data



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Observation summary collector."""
from collections import OrderedDict
from typing import Mapping, Any, Optional, List, Dict

import attr
from attr import Factory
from IPython.display import display, Markdown

from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


# Query definition
@attr.s(auto_attribs=True)
class Observation:
    """
    Observation definition.

    Notes
    -----
    caption : str
        The title and index of the observation. Must
        be unique in the observation set.
    description : Optional[str]
        Text description of the observation.
        (default is None)
    data : Any
        The data to be stored for the observation (e.g.
        a pandas DataFrame). The object should implement
        a useable __repr__ to display correctly.
    data_type : Optional[str]
        The data type of the `data` property
    link : Optional[str]
        Link (usually a document-local link) to the
        originating section of the notebook.
        (default is None)
    score : int
        The risk score associated with the observation.
        (default is 0)
    tags : List[str]
        Optional list of tags.
    additional_properties Dict[str, Any]
        Additional properties not covered
        by core properties.

    """

    caption: str
    data: Any
    description: Optional[str] = None
    data_type: Optional[str] = None
    link: Optional[str] = None
    score: int = 0
    tags: List[str] = Factory(list)
    additional_properties: Dict[str, Any] = Factory(dict)

    @classmethod
    def required_fields(cls) -> List[str]:
        """
        Return required fields for Observation instance.

        Returns
        -------
        List[str]
            List of field names.

        """
        return ["caption", "data"]

    @classmethod
    def all_fields(cls) -> List[str]:
        """
        Return all fields of Observation class.

        Returns
        -------
        List[str]
            List of all field names.

        """
        return [field.name for field in attr.fields(cls)]

    def display(self):
        """Display the observation."""
        display(Markdown(f"### {self.caption}"))
        if self.description:
            display(Markdown(self.description))
        display(Markdown(f"Score: {self.score}"))
        if self.link:
            display(Markdown(f"[Go to details](#{self.link})"))
        if self.tags:
            display(Markdown(f'tags: {", ".join(self.tags)}'))
        display(self.data)
        if self.additional_properties:
            display(Markdown("### Additional Properties"))
            # pylint: disable=no-member
            for key, val in self.additional_properties.items():
                display(Markdown(f"**{key}**: {val}"))
            # pylint: enable=no-member


class Observations:
    """Class to collect and display investigation observations."""

    def __init__(self, observationlist: "Observations" = None):
        """
        Create an observation list.

        Parameters
        ----------
        observationlist : Observations, optional
            Initialize from an existing Observations list
            (the default is None)

        """
        self.observation_list: Dict[str, Observation] = OrderedDict()
        if observationlist is not None:
            self.observation_list.update(observationlist.observations)

    @property
    def observations(self) -> Mapping[str, Observation]:
        """
        Return the current list of Observations.

        Returns
        -------
        Mapping[str, Observation]
            The current ordered dictionary of Observations

        """
        return self.observation_list

    def display_observations(self):
        """Display the current observations using IPython.display."""
        for observation in self.observation_list.values():
            display(observation)

    def add_observation(self, observation: Observation = None, **kwargs):
        """
        Add an observation.

        Add an observation as an Observation instance or as a
        set of keyword parameters (see Observation class for
        acceptable values). Any keyword parameters that are not
        properties of Observation will be stored in the
        Observation.additional_properties dictionary

        Parameters
        ----------
        observation : Observation
            An observation instance.

        Other Parameters
        ----------------
        kwargs : str, Any
            List of key value pairs of the property names
            and values of the Observation to be stored.

        """
        if observation is not None:
            self.observation_list[observation.caption] = observation
        else:
            req_fields = set(Observation.required_fields())
            missing_fields = req_fields.difference(kwargs.keys())
            if missing_fields:
                raise ValueError(
                    "The following fields are required",
                    f"in an Observation: {missing_fields}",
                )

            core_fields = {
                k: v for k, v in kwargs.items() if k in Observation.all_fields()
            }
            new_observation = Observation(**core_fields)
            addl_fields = {
                k: v for k, v in kwargs.items() if k not in Observation.all_fields()
            }
            # pylint: disable=no-member
            new_observation.additional_properties.update(addl_fields)
            self.observation_list[new_observation.caption] = new_observation



#  -------------------------------------------------------------------------
#  Copyright (c) juju4. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for
#  license information.
#  --------------------------------------------------------------------------
"""Sumologic Driver class."""
import re
import time
from datetime import datetime, timedelta
from timeit import default_timer as timer
from typing import Any, Dict, Optional, Tuple, Union

import pandas as pd
from requests.exceptions import ConnectionError as ConnError
from requests.exceptions import HTTPError
from sumologic.sumologic import SumoLogic

from ..._version import VERSION
from ...common.exceptions import (
    MsticpyConnectionError,
    MsticpyNotConnectedError,
    MsticpyUserConfigError,
    MsticpyUserError,
)
from ...common.provider_settings import ProviderSettings, get_provider_settings
from ...common.utility import check_kwargs, export
from .driver_base import DriverBase, QuerySource

__version__ = VERSION
__author__ = "juju4"


SUMOLOGIC_CONNECT_ARGS = {
    "connection_str": "(string) The url endpoint (the default is"
    + " 'https://api.us2.sumologic.com/api').",
    "accessid": "(string) The Sumologic accessid, which is used to "
    + "authenticate on Sumologic instance.",
    "accesskey": "(string) The matching Sumologic accesskey.",
}

_HELP_URI = "https://msticpy.readthedocs.io/en/latest/DataProviders.html"
_SL_NB_URI = (
    "https://github.com/microsoft/msticpy/blob/pr/116-sumologic"
    "-driver-2021-04-21/docs/notebooks/Sumologic-DataConnector.ipynb"
)


@export
class SumologicDriver(DriverBase):
    """Driver to connect and query from Sumologic."""

    _SUMOLOGIC_REQD_ARGS = ["connection_str", "accessid", "accesskey"]
    _CONNECT_DEFAULTS: Dict[str, Any] = {
        "connection_str": "https://api.us2.sumologic.com/api"
    }
    _TIME_FORMAT = '"%Y-%m-%d %H:%M:%S.%6N"'
    _DEF_CHECKINTERVAL = 3
    _DEF_TIMEOUT = 300

    def __init__(self, **kwargs):
        """Instantiate Sumologic Driver."""
        super().__init__(**kwargs)
        self.service = None
        self._loaded = True
        self._connected = False
        self._debug = kwargs.get("debug", False)
        self.public_attribs = {
            "client": self.service,
        }
        self.formatters = {"datetime": self._format_datetime}
        self.checkinterval = self._DEF_CHECKINTERVAL
        self.timeout = self._DEF_TIMEOUT

    def connect(self, connection_str: str = None, **kwargs):
        """
        Connect to Sumologic via sumologic-sdk.

        Parameters
        ----------
        connection_str :
            Sumologic API url endpoint. default: https://api.us2.sumologic.com/api

        Other Parameters
        ----------------
        kwargs :
            Connection parameters can be supplied as keyword parameters.

        Notes
        -----
        Default configuration is read from the DataProviders/Sumologic
        section of msticpyconfig.yaml, if available.

        """
        cs_dict = self._get_connect_args(connection_str, **kwargs)

        arg_dict = {
            key: val for key, val in cs_dict.items() if key in SUMOLOGIC_CONNECT_ARGS
        }
        try:
            # https://github.com/SumoLogic/sumologic-python-sdk/blob/master/scripts/search-job.py
            self.service = SumoLogic(
                accessId=arg_dict["accessid"],
                accessKey=arg_dict["accesskey"],
                endpoint=arg_dict["connection_str"],
            )
        except ConnError as err:
            raise MsticpyConnectionError(
                f"Authentication error connecting to Sumologic: {err}",
                title="Sumologic connection",
                help_uri=_HELP_URI,
                nb_uri=_SL_NB_URI,
            ) from err
        except HTTPError as err:
            raise MsticpyConnectionError(
                f"Communication error connecting to Sumologic: {err}",
                title="Sumologic connection",
                help_uri=_HELP_URI,
                nb_uri=_SL_NB_URI,
            ) from err
        except Exception as err:
            raise MsticpyConnectionError(
                f"Error connecting to Sumologic: {err}",
                title="Sumologic connection",
                help_uri=_HELP_URI,
                nb_uri=_SL_NB_URI,
            ) from err
        self._connected = True
        print(f"connected with accessid {arg_dict['accessid']}")

    def _get_connect_args(
        self, connection_str: Optional[str], **kwargs
    ) -> Dict[str, Any]:
        """Check and consolidate connection parameters."""
        cs_dict: Dict[str, Any] = self._CONNECT_DEFAULTS
        # Fetch any config settings
        cs_dict.update(self._get_config_settings())
        # If a connection string - parse this and add to config
        if connection_str:
            cs_dict["connection_str"] = connection_str
        if kwargs:
            # if connection args supplied as kwargs
            cs_dict.update(kwargs)
            check_kwargs(cs_dict, list(SUMOLOGIC_CONNECT_ARGS.keys()))

        missing_args = set(self._SUMOLOGIC_REQD_ARGS) - cs_dict.keys()
        if missing_args:
            raise MsticpyUserConfigError(
                "One or more connection parameters missing for Sumologic connector",
                ", ".join(missing_args),
                f"Required parameters are {', '.join(self._SUMOLOGIC_REQD_ARGS)}",
                "All parameters:",
                *[f"{arg}: {desc}" for arg, desc in SUMOLOGIC_CONNECT_ARGS.items()],
                title="no Sumologic connection parameters",
                help_uri=_HELP_URI,
                notebook_uri=_SL_NB_URI,
            )
        return cs_dict

    # pylint: disable=broad-except
    def _query(
        self, query: str, query_source: QuerySource = None, **kwargs
    ) -> Union[pd.DataFrame, Any]:
        """
        Execute Sumologic query and retrieve results.

        Parameters
        ----------
        query : str
            Sumologic query to execute
        query_source : QuerySource
            Not used.

        Other Parameters
        ----------------
        kwargs :
            Are passed to Sumologic
            days: Search the past X days.
            start_time: A datetime() object representing the start of the search
                    window. If used without end_time, the end of the search
                    window is the current time.
            end_time: A datetime() object representing the end of the search window.
                  If used without start_time, the search start will be the earliest
                  time in the index.
            timezone: timezone used for time range search
            byreceipttime:  if time reference should used _receiptTime (time when Sumologic
                        got message) instead of _messageTime (time present in log message).
            limit: An integer describing the max number of search results to return.
            forcemessagesresults: Force results to be raw messages even if aggregated query.
            verbosity: Provide more verbose state. from 0 least verbose to 4 most one.
            checkinterval: interval in seconds to check if results are gathered
            timeout: timeout in seconds when gathering results

        Returns
        -------
        Union[pd.DataFrame, Any]
            Query results in a dataframe.
            or query response if an error.

        """
        del query_source
        if not self._connected:
            raise self._create_not_connected_err()

        verbosity = kwargs.pop("verbosity", 0)
        timezone = kwargs.pop("timezone", "UTC")
        by_receipt_time = kwargs.pop("byreceipttime", False)
        self.checkinterval = kwargs.pop("checkinterval", self._DEF_CHECKINTERVAL)
        self.timeout = kwargs.pop("timeout", self._DEF_TIMEOUT)

        start_time, end_time = self._get_time_params(**kwargs)

        # default to unlimited query unless count is specified
        if "limit" in kwargs:
            query = f"{query} | limit {kwargs['limit']}"
        limit = kwargs.pop("limit", 10000)

        if verbosity >= 1:
            print(f"INFO: from {start_time} to {end_time}, TZ {timezone}")

        if verbosity >= 2:
            print(f"DEBUG: query {query}")
            print(f"DEBUG: byreceipttime {by_receipt_time}")
            timer_start = timer()
        # submit the search job
        try:
            searchjob = self.service.search_job(
                query, start_time, end_time, timezone, by_receipt_time
            )
        except Exception as err:
            self._raise_qry_except(err, "submit search_job", "to submit search job")

        qry_count = re.search(r"\|\s*count", query, re.IGNORECASE)
        if verbosity >= 2:
            print(f"DEBUG: search job {searchjob}")
            print(f"DEBUG: search performance: {timer() - timer_start}")
            print(f"DEBUG: messages or records? {qry_count}")

        # poll for job completion
        status = self._poll_job_status(searchjob, verbosity)
        print(status["state"])

        # return the results
        return self._get_job_results(
            searchjob=searchjob,
            status=status,
            qry_count=qry_count,
            force_mssg_rstls=kwargs.pop("forcemessagesresults", False),
            limit=limit,
        )

    def _poll_job_status(self, searchjob, verbosity):
        status = self.service.search_job_status(searchjob)
        if verbosity >= 2:
            print(f"DEBUG: status {status}")
        time_counter = 0
        while status["state"] != "DONE GATHERING RESULTS":
            if status["state"] == "CANCELLED":
                break
            status = self.service.search_job_status(searchjob)
            if verbosity >= 4:
                print(
                    f"DEBUG: pending results, state {status['state']}.",
                    f"slept {time_counter}s. sleeping extra {self.checkinterval}s",
                    "until {self.timeout}s",
                )
            if time_counter < self.timeout:
                time.sleep(self.checkinterval)
                time_counter += self.checkinterval
            else:
                print(
                    f"WARN: wait more than timeout {self.timeout}. stopping. "
                    + "Use timeout argument to wait longer."
                )
                break
        return status

    # pylint: disable=inconsistent-return-statements
    # I don't think there are any - everything returns a list
    def _get_job_results(self, searchjob, status, qry_count, force_mssg_rstls, limit):
        if status["state"] != "DONE GATHERING RESULTS":
            return []
        if not qry_count or force_mssg_rstls:
            # Non-aggregated results, Messages only
            count = status["messageCount"]
            limit2 = (
                count if count < limit and count != 0 else limit
            )  # compensate bad limit check
            try:
                result = self.service.search_job_messages(searchjob, limit=limit2)
                return result["messages"]
            except Exception as err:
                self._raise_qry_except(
                    err, "search_job_messages", "to get job messages"
                )
        else:
            # Aggregated results
            count = status["recordCount"]
            limit2 = (
                count if count < limit and count != 0 else limit
            )  # compensate bad limit check
            try:
                result = self.service.search_job_records(searchjob, limit=limit2)
                return result["records"]
            except Exception as err:
                self._raise_qry_except(
                    err, "search_job_records", "to get search records"
                )

    # pylint: enable=inconsistent-return-statements

    @staticmethod
    def _raise_qry_except(err: Exception, mssg: str, action: Optional[str] = None):
        if isinstance(err, HTTPError):
            raise MsticpyConnectionError(
                f"Communication error connecting to Sumologic: {err}",
                title=f"Sumologic {mssg}",
                help_uri=_HELP_URI,
                notebook_uri=_SL_NB_URI,
            ) from err
        action = action or mssg
        raise MsticpyConnectionError(
            f"Failed {action}: {err}",
            title=f"Sumologic - {mssg}",
            help_uri=_HELP_URI,
            notebook_uri=_SL_NB_URI,
        ) from err

    def _get_time_params(self, **kwargs):
        if "days" in kwargs:
            end = datetime.now()
            start = end - timedelta(days=int(kwargs["days"]))
            return self._format_datetime(start), self._format_datetime(end)

        start = kwargs.pop("start", kwargs.pop("start_time", None))
        end = kwargs.pop("end", kwargs.pop("end_time", None))
        if start and not end:
            end = datetime.now()
        elif not start:
            raise MsticpyUserError(
                "Error! requires either 'days' or 'start' parameters",
                title="Missing parameter.",
                help_uri=_HELP_URI,
                notebook_uri=_SL_NB_URI,
            )
        return self._format_datetime(start), self._format_datetime(end)

    def query(
        self, query: str, query_source: QuerySource = None, **kwargs
    ) -> Union[pd.DataFrame, Any]:
        """
        Execute Sumologic query and retrieve results.

        Parameters
        ----------
        query : str
            Sumologic query to execute
        query_source : QuerySource
            Not used.

        Other Parameters
        ----------------
        kwargs :
            Are passed to Sumologic
            days: Search the past X days.
            start_time (or start): A datetime() object representing the start of the search
                    window. If used without end_time, the end of the search
                    window is the current time.
            end_time (or end): A datetime() object representing the end of the search window.
                  If used without start_time, the search start will be the earliest
                  time in the index.
            timeZone: timezone used for time range search
            byReceiptTime:  if time reference should used _receiptTime (time when Sumologic
                        got message) instead of _messageTime (time present in log message).
            limit: An integer describing the max number of search results to return.
            forceMessagesResults: Force results to be raw messages even if aggregated query.
            verbosity: Provide more verbose state. from 0 least verbose to 4 most one.
            normalize: If set to True, fields containing structures (i.e. subfields)
                   will be flattened such that each field has it's own column in
                   the dataframe. If False, there will be a single column for the
                   structure, with a JSON string encoding all the contents.
            exporting: Export result to file.
            export_path: file path for exporte results.

        Returns
        -------
        Union[pd.DataFrame, Any]
            Query results in a dataframe.
            or query response if an error.

        """
        verbosity = kwargs.get("verbosity", 0)
        normalize = kwargs.pop("normalize", True)
        exporting = kwargs.pop("exporting", False)
        export_path = kwargs.pop("export_path", "")

        results = self._query(query, **kwargs)
        if verbosity >= 3:
            print("DEBUG: {results}")
        if normalize:
            dataframe_res = pd.json_normalize(results)
        else:
            dataframe_res = pd.DataFrame(results)

        for col in dataframe_res.columns:
            if col in ("map._count", "map._timeslice"):
                dataframe_res[col] = pd.to_numeric(dataframe_res[col])

        if exporting:
            if export_path.endswith(".xlsx"):
                if verbosity >= 2:
                    print(f"DEBUG: Exporting results to excel file {export_path}")
                dataframe_res.to_excel(export_path, index=False)
            elif export_path.endswith(".csv"):
                if verbosity >= 2:
                    print("DEBUG: Exporting results to csv file {export_path}")
                dataframe_res.to_csv(export_path, index=False)

        return dataframe_res

    def query_with_results(self, query: str, **kwargs) -> Tuple[pd.DataFrame, Any]:
        """
        Execute query string and return DataFrame of results.

        Parameters
        ----------
        query : str
            Query to execute against Sumologic instance.

        Returns
        -------
        Union[pd.DataFrame,Any]
            A DataFrame (if successful) or
            the underlying provider result if an error occurs.

        """

    # Parameter Formatting methods
    @staticmethod
    def _format_datetime(date_time: datetime) -> str:
        """Return datetime-formatted string."""
        return date_time.strftime("%Y-%m-%dT%H:%M:%S")

    # Read values from configuration
    @staticmethod
    def _get_config_settings() -> Dict[Any, Any]:
        """Get config from msticpyconfig."""
        data_provs = get_provider_settings(config_section="DataProviders")
        sumologic_settings: Optional[ProviderSettings] = data_provs.get("Sumologic")
        return getattr(sumologic_settings, "args", {})

    @staticmethod
    def _create_not_connected_err():
        return MsticpyNotConnectedError(
            "Please run the connect() method before running this method.",
            title="not connected to Sumologic.",
            help_uri=_HELP_URI,
            notebook_uri=_SL_NB_URI,
        )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
syslog_utils - Syslog parsing and utility module.

Functions required to correct collect, parse and visualize syslog data.

Designed to support standard linux syslog for investigations where
auditd is not available.

"""
import datetime as dt
from typing import Any, Dict

import ipywidgets as widgets
import pandas as pd
import pytz

from .._version import VERSION
from ..common.exceptions import MsticpyException
from ..common.utility import export
from ..datamodel.entities import GeoLocation, Host, IpAddress
from .ip_utils import convert_to_ip_entities

__version__ = VERSION
__author__ = "Pete Bryan"

_DETECTIONS_DEF_DIR = "resources"

WIDGET_DEFAULTS = {
    "layout": widgets.Layout(width="95%"),
    "style": {"description_width": "initial"},
}


@export
def create_host_record(
    syslog_df: pd.DataFrame, heartbeat_df: pd.DataFrame, az_net_df: pd.DataFrame = None
) -> Host:
    """
    Generate host_entity record for selected computer.

    Parameters
    ----------
    syslog_df : pd.DataFrame
        A dataframe of all syslog events for the host in the time window requried
    heartbeat_df : pd.DataFrame
        A dataframe of heartbeat data for the host
    az_net_df : pd.DataFrame
        Option dataframe of Azure network data for the host

    Returns
    -------
    Host
        Details of the host data collected

    """
    host_entity = Host(src_event=syslog_df.iloc[0])
    # Produce list of processes on the host that are not
    # part of a 'standard' linux distro
    _apps = syslog_df["ProcessName"].unique().tolist()
    applications = [
        app
        for app in _apps
        if app
        not in (
            "CRON",
            "sudo",
            "snapd",
            "systemd-resolved",
            "systemd",
            "crontab",
            "systemd-timesyncd",
            "systemd-logind",
            "rsyslogd",
            "syslog-ng",
        )
    ]

    # Produce host_entity record mapping linux heartbeat elements to host_entity fields
    if heartbeat_df is not None and not heartbeat_df.empty:
        host_hb = heartbeat_df.iloc[0]
        host_entity.SourceComputerId = host_hb["SourceComputerId"]  # type: ignore
        host_entity.OSType = host_hb["OSType"]  # type: ignore
        host_entity.OSName = host_hb["OSName"]  # type: ignore
        host_entity.OSVMajorersion = host_hb["OSMajorVersion"]  # type: ignore
        host_entity.OSVMinorVersion = host_hb["OSMinorVersion"]  # type: ignore
        host_entity.ComputerEnvironment = host_hb["ComputerEnvironment"]  # type: ignore
        host_entity.OmsSolutions = [  # type: ignore
            sol.strip() for sol in host_hb["Solutions"].split(",")
        ]  # type: ignore
        host_entity.Applications = applications  # type: ignore
        host_entity.VMUUID = host_hb["VMUUID"]  # type: ignore
        ip_entity = IpAddress()
        ip_entity.Address = host_hb["ComputerIP"]
        geoloc_entity = GeoLocation()
        geoloc_entity.CountryName = host_hb["RemoteIPCountry"]  # type: ignore
        geoloc_entity.Longitude = host_hb["RemoteIPLongitude"]  # type: ignore
        geoloc_entity.Latitude = host_hb["RemoteIPLatitude"]  # type: ignore
        ip_entity.Location = geoloc_entity  # type: ignore
        host_entity.IPAddress = ip_entity  # type: ignore

    # If Azure network data present add this to host record
    if az_net_df is not None and not az_net_df.empty:
        if len(az_net_df) == 1:
            priv_addr_str = az_net_df["PrivateIPAddresses"].loc[0]
            host_entity["private_ips"] = convert_to_ip_entities(priv_addr_str)
            pub_addr_str = az_net_df["PublicIPAddresses"].loc[0]
            host_entity["public_ips"] = convert_to_ip_entities(pub_addr_str)
        else:
            if "private_ips" not in host_entity:
                host_entity["private_ips"] = []
            if "public_ips" not in host_entity:
                host_entity["public_ips"] = []

    return host_entity


@export
def cluster_syslog_logons_df(logon_events: pd.DataFrame) -> pd.DataFrame:
    """
    Cluster logon sessions in syslog by start/end time based on PAM events.

    Parameters
    ----------
    logon_events: pd.DataFrame
        A DataFrame of all syslog logon events
        (can be generated with LinuxSyslog.user_logon query)

    Returns
    -------
    logon_sessions: pd.DataFrame
        A dictionary of logon sessions including start and end times
        and logged on user

    Raises
    ------
    MsticpyException
        There are no logon sessions in the supplied data set

    """
    users = []
    starts = []
    ends = []
    ses_close_time = logon_events["TimeGenerated"].max()
    ses_opened = 0
    ses_closed = 0
    # Extract logon session opened and logon session closed data.
    logons_opened = (
        (
            logon_events[
                logon_events["SyslogMessage"].str.contains("pam_unix.+session opened")
            ]
        )
        .set_index("TimeGenerated")
        .sort_index(ascending=True)
    )
    logons_closed = (
        (
            logon_events[
                logon_events["SyslogMessage"].str.contains("pam_unix.+session closed")
            ]
        )
        .set_index("TimeGenerated")
        .sort_index(ascending=True)
    )
    if logons_opened.empty or logons_closed.empty:
        raise MsticpyException("There are no logon sessions in the supplied data set")

    # For each session identify the likely start and end times
    while ses_opened < len(logons_opened.index) and ses_closed < len(
        logons_closed.index
    ):
        ses_start = (logons_opened.iloc[ses_opened]).name
        ses_end = (logons_closed.iloc[ses_closed]).name
        # If we can identify a user for the session add this to the details
        if "User" in logons_opened.columns:
            user = (logons_opened.iloc[ses_opened]).User
        elif "Sudoer" in logons_opened.columns:
            user = (logons_opened.iloc[ses_opened]).Sudoer
        else:
            user = "Unknown"
        if ses_start <= ses_close_time and ses_opened != 0:
            ses_opened += 1
            continue
        if ses_end < ses_start:
            ses_closed += 1
            continue
        users.append(user)
        starts.append(ses_start)
        ends.append(ses_end)
        ses_close_time = ses_end
        ses_closed += 1
        ses_opened += 1
    return pd.DataFrame({"User": users, "Start": starts, "End": ends})


@export
def risky_sudo_sessions(
    sudo_sessions: pd.DataFrame,
    risky_actions: dict = None,
    suspicious_actions: list = None,
) -> dict:
    """
    Detect if a sudo session occurs at the point of a suspicious event.

    Parameters
    ----------
    sudo_sessions: dict
        Dictionary of sudo sessions (as generated by cluster_syslog_logons)
    risky_actions: dict (Optional)
        Dictionary of risky sudo commands (as generated by cmd_line.risky_cmd_line)
    suspicious_actions: list (Optional)
        List of risky sudo commands (as generated by cmd_line.cmd_speed)

    Returns
    -------
    risky_sessions: dict
        A dictionary of sudo sessions with flags denoting risk

    """
    sessions = sudo_sessions[["User", "Start", "End"]].to_dict("index")

    if risky_actions is None and suspicious_actions is None:
        raise MsticpyException(
            "At least one of risky_actions or suspicious_actions must be supplied"
        )

    # Depending on whether we have risky or suspicious acitons or both
    # identify sessions which these actions occur in
    risky_act_sessions: Dict[str, Any] = {}
    susp_act_sessions: Dict[str, Any] = {}
    if risky_actions is not None:
        risky_act_sessions = _find_risky_sudo_session(
            risky_actions=risky_actions, sudo_sessions=sessions
        )
    if suspicious_actions is not None:
        susp_act_sessions = _find_suspicious_sudo_session(
            suspicious_actions=suspicious_actions, sudo_sessions=sessions
        )
    return {**risky_act_sessions, **susp_act_sessions}


def _normalize_to_utc(time_stamp: dt.datetime):
    # Normalize datetimes to UTC in case we have mixed timezones in datasets
    if time_stamp.tzinfo is None or time_stamp.tzinfo.utcoffset(time_stamp) is None:
        time_stamp = time_stamp.replace(tzinfo=pytz.UTC)
    else:
        time_stamp = time_stamp.astimezone(pytz.utc)
    return time_stamp


def _find_risky_sudo_session(risky_actions: dict, sudo_sessions: dict):
    risky_sessions = {}
    # Determine if risky event occurs during a session time window
    for key, value in risky_actions.items():
        for sess_key, sess_val in sudo_sessions.items():
            if (
                _normalize_to_utc(sess_val["Start"])
                <= _normalize_to_utc(key)
                <= _normalize_to_utc(sess_val["End"])
            ):
                risky_sessions.update({sess_key: value})
    return risky_sessions


def _find_suspicious_sudo_session(suspicious_actions: list, sudo_sessions: dict):
    risky_sessions = {}
    # Determine if suspicious event occurs during a session time window
    for event in suspicious_actions:
        for value in event.values():
            for sess_key, sess_val in sudo_sessions.items():
                if (
                    _normalize_to_utc(sess_val["Start"])
                    <= _normalize_to_utc(value[0]["TimeGenerated"].iloc[1])
                    <= _normalize_to_utc(sess_val["End"])
                ):
                    risky_sessions.update({sess_key: "Suspicious event pattern"})
    return risky_sessions



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Setup script for msticpy."""
import os
import re
import setuptools


def install_requires_rtd(install_list: list) -> list:
    """Return modified install list if installing for ReadtheDocs."""
    rtd_exceptions = [
        "Kqlmagic",
        "azure-cli-core",
        "matplotlib",
        "statsmodels",
        "scipy",
        "splunk-sdk",
        "seaborn",
    ]
    return [
        pkg
        for pkg in install_list
        if not any(excl_pkg for excl_pkg in rtd_exceptions if excl_pkg in pkg)
    ]


with open("README.md", "r") as fh:
    LONG_DESC = fh.read()

# pylint: disable=locally-disabled, invalid-name
with open("msticpy/_version.py", "r") as fd:
    v_match = re.search(r'^VERSION\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE)
    __version__ = v_match.group(1) if v_match else "no version"
# pylint: enable=locally-disabled, invalid-name

with open("requirements.txt", "r") as fh:
    INSTALL_REQUIRES = fh.readlines()

with open("requirements-dev.txt", "r") as fh:
    INSTALL_DEV_REQUIRES = fh.readlines()


def _combine_extras(extras: list) -> list:
    return list(
        {pkg for name, pkgs in EXTRAS.items() for pkg in pkgs if name in extras}
    )


# Extras definitions
EXTRAS = {
    "dev": INSTALL_DEV_REQUIRES,
    "vt3": ["vt-py>=0.6.1", "vt-graph-api>=1.0.1", "nest_asyncio>=1.4.0"],
    "splunk": ["splunk-sdk>=1.6.0"],
    "sumologic": ["sumologic-sdk>=0.1.11", "openpyxl>=3.0"],
    "kql": ["KqlmagicCustom[jupyter-extended]>=0.1.114.dev25"],
    "_azure_core": [
        "azure-mgmt-compute>=4.6.2",
        "azure-mgmt-core>=1.2.1",
        "azure-mgmt-monitor>=2.0.0",
        "azure-mgmt-network>=2.7.0",
        "azure-mgmt-resource>=16.1.0",
        "azure-storage-blob>=12.5.0",
        "azure-mgmt-resourcegraph>=8.0.0",
    ],
    "keyvault": [
        "azure-keyvault-secrets>=4.0.0",
        "azure-mgmt-keyvault>=2.0.0",
        "keyring>=13.2.1",  # needed by Key Vault package
    ],
    "ml": ["scikit-learn>=0.20.2", "scipy>=1.1.0", "statsmodels>=0.11.1"],
    "sql2kql": ["moz_sql_parser>=4.5.0,<=4.11.21016"],
    "riskiq": ["passivetotal>=2.5.3"],
}
extras_all = [
    extra for name, extras in EXTRAS.items() for extra in extras if name != "dev"
]
EXTRAS["all"] = extras_all

# Create combination extras
EXTRAS["all"] = sorted(
    _combine_extras(list({name for name in EXTRAS if name != "dev"}))
)

EXTRAS["azure"] = sorted(_combine_extras(["_azure_core", "keyvault"]))
EXTRAS["test"] = sorted(_combine_extras(["all", "dev"]))
EXTRAS["azsentinel"] = sorted(_combine_extras(["azure", "kql", "keyvault"]))
EXTRAS["azuresentinel"] = sorted(_combine_extras(["azure", "kql", "keyvault"]))

# If ReadTheDocs build, remove a couple of problematic packages
# (we ask Sphinx to mock these in the import)
if os.environ.get("MP_RTD_BUILD"):
    INSTALL_REQUIRES = install_requires_rtd(INSTALL_REQUIRES)

setuptools.setup(
    name="msticpy",
    version=__version__,
    author="Ian Hellen",
    author_email="ianhelle@microsoft.com",
    description="MSTIC Security Tools",
    license="MIT License",
    long_description=LONG_DESC,
    long_description_content_type="text/markdown",
    url="https://github.com/microsoft/msticpy",
    project_urls={
        "Documentation": "https://msticpy.readthedocs.io",
        "Code": "https://github.com/microsoft/msticpy",
    },
    python_requires=">=3.6",
    packages=setuptools.find_packages(exclude=["tests", "tests.*", "*.tests.*"]),
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 5 - Production/Stable",
    ],
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS,
    keywords=[
        "security",
        "azure",
        "sentinel",
        "mstic",
        "cybersec",
        "infosec",
        "cyber",
        "cybersecurity",
        "jupyter",
        "notebooks",
        "SOC",
        "hunting",
    ],
    zip_safe=False,
    include_package_data=True,
)
